#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        if gridinfo.tile_type in ['CLBLL_L', 'CLBLL_R', 'CLBLM_L', 'CLBLM_R']:
            site_name = sorted(gridinfo.sites.keys())[0]
            yield tile_name, site_name


def write_params(params):
    pinstr = 'tile,val,site\n'
    for tile, (site, val) in sorted(params.items()):
        pinstr += '%s,%s,%s\n' % (tile, val, site)
    open('params.csv', 'w').write(pinstr)


def run():
    print(
        '''
module top(input clk, stb, di, output do);
    localparam integer DIN_N = 8;
    localparam integer DOUT_N = 8;

    reg [DIN_N-1:0] din;
    wire [DOUT_N-1:0] dout;

    reg [DIN_N-1:0] din_shr;
    reg [DOUT_N-1:0] dout_shr;

    always @(posedge clk) begin
        din_shr <= {din_shr, di};
        dout_shr <= {dout_shr, din_shr[DIN_N-1]};
        if (stb) begin
            din <= din_shr;
            dout_shr <= dout;
        end
    end

    assign do = dout_shr[DOUT_N-1];
    ''')

    params = {}

    sites = list(gen_sites())
    for (tile_name, site_name), isone in zip(sites,
                                             util.gen_fuzz_states(len(sites))):
        params[tile_name] = (site_name, isone)

    print("endmodule")
    write_params(params)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import sys
import glob
import os
import re
import difflib


def get_file_pairs():
    pairs_list = list()
    for path1 in glob.glob('*.bits'):
        for path2 in glob.glob('*.bits'):
            file1 = os.path.basename(path1)
            file2 = os.path.basename(path2)
            if file1 == file2:
                continue
            files_pair = [file1, file2]
            files_pair.sort()
            pairs_list.append(files_pair[0] + ":" + files_pair[1])
    pairs_set = set(pairs_list)
    for pair in pairs_set:
        file1, file2 = pair.split(":")
        yield file1, file2


def extract_parameters_string(basename):
    params_str = re.search('^design_(.*).bits$', basename)
    return params_str.group(1)


def extract_parameters(basename):
    iostandard, slew, drive = extract_parameters_string(basename).split('_')
    return iostandard, slew, drive


def generate_differing_bits(basename1, basename2):
    with open(basename1, 'r') as path1:
        with open(basename2, 'r') as path2:
            diff = difflib.unified_diff(
                path1.read().splitlines(),
                path2.read().splitlines(),
                fromfile='path1',
                tofile='path2')
            for line in diff:
                if line.startswith('---'):
                    continue
                if line.startswith('+++'):
                    continue
                if line.startswith('@'):
                    continue
                if line.startswith('-'):
                    yield extract_parameters_string(basename1), line.strip('-')
                    continue
                if line.startswith('+'):
                    yield extract_parameters_string(basename2), line.strip('+')
                    continue


class Database():
    def __init__(self, convert_bits=False):
        self.all_bits = set()
        self.properties_bits = dict()
        self.convert_bits = convert_bits
        self.populate()

    def populate(self):
        for file1, file2 in get_file_pairs():
            #print(file1 + " vs " + file2)
            for property_str, bit in generate_differing_bits(file1, file2):
                #print(property_str + " " + bit)
                self.update_all_bits(bit)
                if property_str in self.properties_bits:
                    self.properties_bits[property_str].add(bit)
                else:
                    self.properties_bits[property_str] = set()
                    self.properties_bits[property_str].add(bit)

    def update_all_bits(self, bit):
        self.all_bits.add(bit)

    def print_all_bits(self):
        print(self.all_bits)

    def get_keys(self):
        return self.properties_bits.keys()

    def print_bits(self, key):
        if key in self.properties_bits:
            print("%s: %s" % (key, self.properties_bits[key]))
        else:
            print("The specified property is not in the database")

    def convert_bit_format(self, item):
        dummy, address, word, bit = item.split("_")
        address = int(address[-2:], 16)
        bit = int(word) % 4 * 32 + int(bit)
        return "{address}_{bit}".format(address=address, bit=bit)

    def convert_header(self, header):
        converted_bits = []
        for bit in header:
            #print(bit + ":" + self.convert_bit_format(bit))
            converted_bits.append(self.convert_bit_format(bit))
        return converted_bits

    def get_csv_header(self):
        header = list(self.all_bits)
        header.sort()
        self.csv_header = header
        if self.convert_bits:
            header = self.convert_header(header)
        line = "property,v,i,r,"
        for title in header:
            line += title + ","
        return line + '\n'

    def extract_rvi_parameters(self, rvi):
        iostandard, slew, drive = rvi.split("_")
        if iostandard[-2:] == "12":
            voltage = 1.2
        elif iostandard[-2:] == "15":
            voltage = 1.5
        elif iostandard[-2:] == "18":
            voltage = 1.8
        elif iostandard[-2:] == "25":
            voltage = 2.5
        else:
            voltage = 3.3
        resistance = voltage / (int(drive) * 0.001)
        return "%.1f,%s,%.3f" % (voltage, drive, resistance)

    def get_csv_body(self):
        lines = ""
        keys = list(self.get_keys())
        keys.sort()
        for properties_key in keys:
            line = properties_key + "," + self.extract_rvi_parameters(
                properties_key) + ","
            for title in self.csv_header:
                if title in self.properties_bits[properties_key]:
                    line += "X,"
                else:
                    line += " ,"
            line += '\n'
            lines += line
        return lines

    def write_csv(self, filename):
        filename = os.getcwd() + "/" + filename
        fp = open(filename, 'w')
        fp.write(self.get_csv_header())
        fp.write(self.get_csv_body())
        fp.close()
        print("Written results to %s file.\n" % filename)


def main():
    database = Database(True)
    database.write_csv("differences.csv")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Reduce sites types to prototypes that are always correct.

reduce_tile_types.py generates per tile type site types.  reduce_site_types.py
takes all site types across all tiles and creates generic site types that are
valid for all tile types.

"""

import argparse
import prjxray.lib
import os
import os.path
import re
import json


def main():
    parser = argparse.ArgumentParser(
        description="Reduces per tile site types to generic site types.")
    parser.add_argument('--output_dir', required=True)

    args = parser.parse_args()

    SITE_TYPE = re.compile('^tile_type_(.+)_site_type_(.+)\.json$')
    site_types = {}
    for path in os.listdir(args.output_dir):
        match = SITE_TYPE.fullmatch(path)
        if match is None:
            continue

        site_type = match.group(2)
        if site_type not in site_types:
            site_types[site_type] = []

        site_types[site_type].append(path)

    for site_type in site_types:
        proto_site_type = None
        for instance in site_types[site_type]:
            with open(os.path.join(args.output_dir, instance)) as f:
                instance_site_type = json.load(f)

                for site_pin in instance_site_type['site_pins'].values():
                    if 'index_in_site' in site_pin:
                        del site_pin['index_in_site']

            if proto_site_type is None:
                proto_site_type = instance_site_type
            else:
                prjxray.lib.compare_prototype_site(
                    proto_site_type,
                    instance_site_type,
                )

        with open(os.path.join(args.output_dir,
                               'site_type_{}.json'.format(site_type)),
                  'w') as f:
            json.dump(proto_site_type, f)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import copy
import json
import os
from utils import xjson
'''
Historically we grouped data into "segments"
These were a region of the bitstream that encoded one or more tiles
However, this didn't scale with certain tiles like BRAM
Some sites had multiple bitstream areas and also occupied multiple tiles

Decoding was then shifted to instead describe how each title is encoded
A post processing step verifies that two tiles don't reference the same bitstream area
'''

import util as localutil


def nolr(tile_type):
    '''
    Remove _L or _R suffix tile_type suffix, if present
    Ex: BRAM_INT_INTERFACE_L => BRAM_INT_INTERFACE
    Ex: VBRK => VBRK
    '''
    postfix = tile_type[-2:]
    if postfix in ('_L', '_R'):
        return tile_type[:-2]
    else:
        return tile_type


def make_tiles_by_grid(database):
    # lookup tile names by (X, Y)
    tiles_by_grid = dict()

    for tile_name in database:
        tile = database[tile_name]
        tiles_by_grid[(tile["grid_x"], tile["grid_y"])] = tile_name

    return tiles_by_grid


def propagate_INT_lr_bits(
        database, tiles_by_grid, tile_frames_map, verbose=False):
    '''Populate segment base addresses: L/R along INT column'''

    int_frames, int_words, _ = localutil.get_entry('INT', 'CLB_IO_CLK')

    verbose and print('')
    for tile in database:
        if database[tile]["type"] not in ["INT_L", "INT_R"]:
            continue

        if not database[tile]["bits"]:
            continue

        grid_x = database[tile]["grid_x"]
        grid_y = database[tile]["grid_y"]
        baseaddr = int(database[tile]["bits"]["CLB_IO_CLK"]["baseaddr"], 0)
        offset = database[tile]["bits"]["CLB_IO_CLK"]["offset"]

        if database[tile]["type"] == "INT_L":
            grid_x += 1
            baseaddr = baseaddr + 0x80
        elif database[tile]["type"] == "INT_R":
            grid_x -= 1
            baseaddr = baseaddr - 0x80
        else:
            assert 0, database[tile]["type"]

        # ROI at edge?
        if (grid_x, grid_y) not in tiles_by_grid:
            verbose and print('  Skip edge')
            continue

        other_tile = tiles_by_grid[(grid_x, grid_y)]

        if database[tile]["type"] == "INT_L":
            assert database[other_tile]["type"] == "INT_R"
        elif database[tile]["type"] == "INT_R":
            assert database[other_tile]["type"] == "INT_L"
        else:
            assert 0

        localutil.add_tile_bits(
            other_tile, database[other_tile], baseaddr, offset, int_frames,
            int_words, tile_frames_map)


def propagate_INT_bits_in_column(database, tiles_by_grid, tile_frames_map):
    """ Propigate INT offsets up and down INT columns.

    INT columns appear to be fairly regular, where starting from offset 0,
    INT tiles next to INT tiles increase the word offset by 2.  The HCLK tile
    is surrounded above and sometimes below by an INT tile.  Because the HCLK
    tile only useds one word, the offset increase by one at the HCLK.

    """

    seen_int = set()

    int_frames, int_words, _ = localutil.get_entry('INT', 'CLB_IO_CLK')
    hclk_frames, hclk_words, _ = localutil.get_entry('HCLK', 'CLB_IO_CLK')

    for tile_name in sorted(database.keys()):
        tile = database[tile_name]

        if tile['type'] not in ['INT_L', 'INT_R']:
            continue

        l_or_r = tile['type'][-1]

        if not tile['bits']:
            continue

        if tile_name in seen_int:
            continue

        # Walk down INT column
        while True:
            seen_int.add(tile_name)

            next_tile = tiles_by_grid[(tile['grid_x'], tile['grid_y'] + 1)]
            next_tile_type = database[next_tile]['type']

            if tile['bits']['CLB_IO_CLK']['offset'] == 0:
                assert next_tile_type in [
                    'B_TERM_INT', 'BRKH_INT', 'BRKH_B_TERM_INT'
                ], next_tile_type
                break

            baseaddr = int(tile['bits']['CLB_IO_CLK']['baseaddr'], 0)
            offset = tile['bits']['CLB_IO_CLK']['offset']

            if tile['type'].startswith(
                    'INT_') and next_tile_type == tile['type']:
                # INT next to INT
                offset -= int_words
                localutil.add_tile_bits(
                    next_tile, database[next_tile], baseaddr, offset,
                    int_frames, int_words, tile_frames_map)
            elif tile['type'].startswith('INT_'):
                # INT above HCLK
                assert next_tile_type.startswith(
                    'HCLK_{}'.format(l_or_r)), next_tile_type

                offset -= hclk_words
                localutil.add_tile_bits(
                    next_tile, database[next_tile], baseaddr, offset,
                    hclk_frames, hclk_words, tile_frames_map)
            else:
                # HCLK above INT
                assert tile['type'].startswith(
                    'HCLK_{}'.format(l_or_r)), tile['type']
                if next_tile_type == 'INT_{}'.format(l_or_r):
                    offset -= int_words
                    localutil.add_tile_bits(
                        next_tile, database[next_tile], baseaddr, offset,
                        int_frames, int_words, tile_frames_map)
                else:
                    # Handle special case column where the PCIE tile is present.
                    assert next_tile_type in ['PCIE_NULL'], next_tile_type
                    break

            tile_name = next_tile
            tile = database[tile_name]

        # Walk up INT column
        while True:
            seen_int.add(tile_name)

            next_tile = tiles_by_grid[(tile['grid_x'], tile['grid_y'] - 1)]
            next_tile_type = database[next_tile]['type']

            if tile['bits']['CLB_IO_CLK']['offset'] == 99:
                assert next_tile_type in [
                    'T_TERM_INT', 'BRKH_INT', 'BRKH_TERM_INT', 'BRKH_INT_PSS'
                ], next_tile_type
                break

            baseaddr = int(tile['bits']['CLB_IO_CLK']['baseaddr'], 0)
            offset = tile['bits']['CLB_IO_CLK']['offset']

            if tile['type'].startswith(
                    'INT_') and next_tile_type == tile['type']:
                # INT next to INT
                offset += int_words
                localutil.add_tile_bits(
                    next_tile, database[next_tile], baseaddr, offset,
                    int_frames, int_words, tile_frames_map)
            elif tile['type'].startswith('INT_'):
                # INT below HCLK
                assert next_tile_type.startswith(
                    'HCLK_{}'.format(l_or_r)), next_tile_type

                offset += int_words
                localutil.add_tile_bits(
                    next_tile, database[next_tile], baseaddr, offset,
                    hclk_frames, hclk_words, tile_frames_map)
            else:
                # HCLK below INT
                assert tile['type'].startswith(
                    'HCLK_{}'.format(l_or_r)), tile['type']
                assert next_tile_type == 'INT_{}'.format(
                    l_or_r), next_tile_type

                offset += hclk_words
                localutil.add_tile_bits(
                    next_tile, database[next_tile], baseaddr, offset,
                    int_frames, int_words, tile_frames_map)

            tile_name = next_tile
            tile = database[tile_name]


def propagate_INT_INTERFACE_bits_in_column(
        database, tiles_by_grid, int_interface_name, tile_frames_map):
    """ Propagate INT_INTERFACE column for a given INT_INTERFACE tile name.

    INT_INTERFACE tiles do not usually have any PIPs or baseaddresses,
    except for a few cases such as PCIE or GTP INTERFACE tiles.

    These are very regular tiles, except for the horizontal clock line,
    which adds a one-word offset.

    This function replicates the baseaddress and calculates the correct offset
    for each INT INTERFACE tile in a column, starting from a tile in the column
    that has the baseaddress calculated from the corresponding tilegrid fuzzer.
    """

    seen_int = set()

    int_frames, int_words, _ = localutil.get_entry('INT', 'CLB_IO_CLK')
    hclk_frames, hclk_words, _ = localutil.get_entry('HCLK', 'CLB_IO_CLK')

    for tile_name in sorted(database.keys()):
        tile = database[tile_name]

        if not tile['type'].startswith(int_interface_name):
            continue

        if not tile['bits']:
            continue

        if tile_name in seen_int:
            continue

        # Walk down INT column
        down_tile = tile
        down_tile_name = tile_name
        while True:
            seen_int.add(down_tile_name)

            baseaddr = int(down_tile['bits']['CLB_IO_CLK']['baseaddr'], 0)
            offset = down_tile['bits']['CLB_IO_CLK']['offset']
            extra_offset = 0

            next_tile = tiles_by_grid[(
                down_tile['grid_x'], down_tile['grid_y'] + 1)]
            if next_tile.startswith("HCLK"):
                next_tile = tiles_by_grid[(
                    down_tile['grid_x'], down_tile['grid_y'] + 2)]
                extra_offset = hclk_words

            next_tile_type = database[next_tile]['type']

            if next_tile_type != tile['type']:
                break

            if next_tile_type == down_tile['type']:
                # INT next to INT
                offset -= (int_words + extra_offset)
                localutil.add_tile_bits(
                    next_tile, database[next_tile], baseaddr, offset,
                    int_frames, int_words, tile_frames_map)

            down_tile_name = next_tile
            down_tile = database[down_tile_name]

        # Walk up INT column
        up_tile = tile
        up_tile_name = tile_name
        while True:
            seen_int.add(up_tile_name)

            baseaddr = int(up_tile['bits']['CLB_IO_CLK']['baseaddr'], 0)
            offset = up_tile['bits']['CLB_IO_CLK']['offset']
            extra_offset = 0

            next_tile = tiles_by_grid[(
                up_tile['grid_x'], up_tile['grid_y'] - 1)]
            if next_tile.startswith("HCLK"):
                next_tile = tiles_by_grid[(
                    up_tile['grid_x'], up_tile['grid_y'] - 2)]
                extra_offset = hclk_words

            next_tile_type = database[next_tile]['type']

            if next_tile_type != tile['type']:
                break

            if next_tile_type == up_tile['type']:
                # INT next to INT
                offset += (int_words + extra_offset)
                localutil.add_tile_bits(
                    next_tile, database[next_tile], baseaddr, offset,
                    int_frames, int_words, tile_frames_map)

            up_tile_name = next_tile
            up_tile = database[up_tile_name]


def propagate_rebuf(database, tiles_by_grid):
    """ Writing a fuzzer for the CLK_BUFG_REBUF tiles is hard, so propigate from CLK_HROW tiles.

    In the clock column, there is a CLK_BUFG_REBUF above and below the CLK_HROW
    tile.  Each clock column appears to use the same offsets, so propigate
    the base address and frame count, and update the offset and word count.

    """
    for tile_name in sorted(database.keys()):
        tile = database[tile_name]

        if tile['type'] not in ['CLK_HROW_BOT_R', 'CLK_HROW_TOP_R']:
            continue

        rebuf_below = tiles_by_grid[(tile['grid_x'], tile['grid_y'] - 12)]
        assert database[rebuf_below]['type'] == 'CLK_BUFG_REBUF', database[
            rebuf_below]['type']
        rebuf_above = tiles_by_grid[(tile['grid_x'], tile['grid_y'] + 13)]
        assert database[rebuf_above]['type'] == 'CLK_BUFG_REBUF', database[
            rebuf_below]['type']

        assert database[tile_name]['bits']['CLB_IO_CLK'][
            'offset'] == 42, database[tile_name]['bits']['CLB_IO_CLK']
        database[rebuf_below]['bits'] = copy.deepcopy(
            database[tile_name]['bits'])
        database[rebuf_below]['bits']['CLB_IO_CLK']['offset'] = 73
        database[rebuf_below]['bits']['CLB_IO_CLK']['words'] = 4

        database[rebuf_above]['bits'] = copy.deepcopy(
            database[tile_name]['bits'])
        database[rebuf_above]['bits']['CLB_IO_CLK']['offset'] = 24
        database[rebuf_above]['bits']['CLB_IO_CLK']['words'] = 4


def propagate_IOB_SING(database, tiles_by_grid):
    """ The IOB_SING are half tiles at top and bottom of every IO column.

    Unlike most tiles, they do not behave consistently.  The tile at the top
    of the column is the bottom half of a full IOB, and the tile at the bottom
    of the column is the top half of a full IOB.  For this reason, explicit
    bit aliasing is used to map the full IOB bits into the two halves, and a
    mapping is provided for the site naming.

    """

    seen_iobs = set()
    for tile in database:
        if tile in seen_iobs:
            continue

        if database[tile]["type"] not in ["LIOB33", "RIOB33", "RIOB18"]:
            continue

        while True:
            prev_tile = tile
            tile = tiles_by_grid[(
                database[tile]['grid_x'], database[tile]['grid_y'] + 1)]
            if '_SING' in database[tile]['type']:
                break

        bottom_tile = tile
        seen_iobs.add(bottom_tile)

        bits = database[prev_tile]['bits']['CLB_IO_CLK']

        while True:
            tile = tiles_by_grid[(
                database[tile]['grid_x'], database[tile]['grid_y'] - 1)]
            seen_iobs.add(tile)

            if '_SING' in database[tile]['type']:
                break

            if 'CLB_IO_CLK' in database[tile]['bits']:
                assert bits['baseaddr'] == database[tile]['bits'][
                    'CLB_IO_CLK']['baseaddr']
                assert bits['frames'] == database[tile]['bits']['CLB_IO_CLK'][
                    'frames']
                assert bits['words'] == database[tile]['bits']['CLB_IO_CLK'][
                    'words']

        top_tile = tile

        database[top_tile]['bits']['CLB_IO_CLK'] = copy.deepcopy(bits)
        database[top_tile]['bits']['CLB_IO_CLK']['words'] = 2
        database[top_tile]['bits']['CLB_IO_CLK']['offset'] = 99
        database[top_tile]['bits']['CLB_IO_CLK']['alias'] = {
            'type': database[prev_tile]['type'],
            'start_offset': 0,
            'sites': {
                'IOB33_Y0': 'IOB33_Y1',
            }
        }

        database[bottom_tile]['bits']['CLB_IO_CLK'] = copy.deepcopy(bits)
        database[bottom_tile]['bits']['CLB_IO_CLK']['words'] = 2
        database[bottom_tile]['bits']['CLB_IO_CLK']['offset'] = 0
        database[bottom_tile]['bits']['CLB_IO_CLK']['alias'] = {
            'type': database[prev_tile]['type'],
            'start_offset': 2,
            'sites': {
                'IOB33_Y0': 'IOB33_Y0',
            }
        }


def propagate_IOI_SING(database, tiles_by_grid):
    """
    The IOI_SING, similarly to IOB_SING, are half tiles at top and bottom of every
    IO column.

    The tile contains half of the sites that are present in the full IOI,
    namely one ILOGIC, OLOGIC and IDELAY.
    """

    seen_iois = set()
    for tile in database:
        if tile in seen_iois:
            continue

        if database[tile]["type"] not in ["LIOI3", "RIOI3", "RIOI"]:
            continue

        while True:
            prev_tile = tile
            tile = tiles_by_grid[(
                database[tile]['grid_x'], database[tile]['grid_y'] + 1)]
            if '_SING' in database[tile]['type']:
                break

        bottom_tile = tile
        seen_iois.add(bottom_tile)

        bits = database[prev_tile]['bits']['CLB_IO_CLK']

        while True:
            tile = tiles_by_grid[(
                database[tile]['grid_x'], database[tile]['grid_y'] - 1)]
            seen_iois.add(tile)

            if '_SING' in database[tile]['type']:
                break

            if 'CLB_IO_CLK' in database[tile]['bits']:
                if tile.startswith("LIOI") or tile.startswith("RIOI"):
                    assert bits['baseaddr'] == database[tile]['bits'][
                        'CLB_IO_CLK']['baseaddr']
                    assert bits['frames'] == database[tile]['bits'][
                        'CLB_IO_CLK']['frames'], "{}:{} == {}".format(
                            tile, bits['frames'],
                            database[tile]['bits']['CLB_IO_CLK']['frames'])
                    assert bits['words'] == database[tile]['bits'][
                        'CLB_IO_CLK']['words'], "{}: {} != {}".format(
                            tile, bits['words'],
                            database[tile]['bits']['CLB_IO_CLK']['words'])

        top_tile = tile

        database[top_tile]['bits']['CLB_IO_CLK'] = copy.deepcopy(bits)
        database[top_tile]['bits']['CLB_IO_CLK']['words'] = 2
        database[top_tile]['bits']['CLB_IO_CLK']['offset'] = 99
        database[top_tile]['bits']['CLB_IO_CLK']['alias'] = {
            'type': database[prev_tile]['type'],
            'start_offset': 0,
            'sites': {}
        }

        database[bottom_tile]['bits']['CLB_IO_CLK'] = copy.deepcopy(bits)
        database[bottom_tile]['bits']['CLB_IO_CLK']['words'] = 2
        database[bottom_tile]['bits']['CLB_IO_CLK']['offset'] = 0
        database[bottom_tile]['bits']['CLB_IO_CLK']['alias'] = {
            'type': database[prev_tile]['type'],
            'start_offset': 2,
            'sites': {}
        }


def propagate_IOI_Y9(database, tiles_by_grid):
    """
    There are IOI tiles (X0Y9 and X43Y9) that have the frame address 1 frame
    higher than the rest, just like for some of the SING tiles.

    """
    ioi_tiles = os.getenv('XRAY_IOI3_TILES')

    assert ioi_tiles is not None, "XRAY_IOI3_TILES env variable not set"
    tiles = ioi_tiles.split(" ")

    for tile in tiles:
        prev_tile = tiles_by_grid[(
            database[tile]['grid_x'], database[tile]['grid_y'] - 1)]
        while database[prev_tile]["type"] != database[tile]["type"]:
            prev_tile = tiles_by_grid[(
                database[prev_tile]['grid_x'],
                database[prev_tile]['grid_y'] - 1)]
        bits = database[prev_tile]['bits']['CLB_IO_CLK']
        database[tile]['bits']['CLB_IO_CLK'] = copy.deepcopy(bits)
        database[tile]['bits']['CLB_IO_CLK']['words'] = 4
        database[tile]['bits']['CLB_IO_CLK']['offset'] = 18


def alias_HCLKs(database):
    """ Generate HCLK aliases for HCLK_[LR] subsets.

    There are some HCLK_[LR] tiles that are missing some routing due to
    obstructions, e.g. PCIE hardblock.  These tiles do not have southbound
    clock routing, but are otherwise the same as HCLK_[LR] tiles.

    Simply alias their segbits.

    """
    for tile in database:
        if database[tile]['type'] == "HCLK_L_BOT_UTURN":
            database[tile]['bits']['CLB_IO_CLK']['alias'] = {
                "sites": {},
                "start_offset": 0,
                "type": "HCLK_L"
            }
        elif database[tile]['type'] == "HCLK_R_BOT_UTURN":
            database[tile]['bits']['CLB_IO_CLK']['alias'] = {
                "sites": {},
                "start_offset": 0,
                "type": "HCLK_R"
            }


def run(json_in_fn, json_out_fn, verbose=False):
    # Load input files
    database = json.load(open(json_in_fn, "r"))
    tiles_by_grid = make_tiles_by_grid(database)

    tile_frames_map = localutil.TileFrames()
    propagate_INT_lr_bits(
        database, tiles_by_grid, tile_frames_map, verbose=verbose)
    propagate_INT_bits_in_column(database, tiles_by_grid, tile_frames_map)
    propagate_INT_INTERFACE_bits_in_column(
        database, tiles_by_grid, "GTP_INT_INTERFACE", tile_frames_map)
    propagate_INT_INTERFACE_bits_in_column(
        database, tiles_by_grid, "GTX_INT_INTERFACE", tile_frames_map)
    propagate_INT_INTERFACE_bits_in_column(
        database, tiles_by_grid, "PCIE_INT_INTERFACE", tile_frames_map)
    propagate_rebuf(database, tiles_by_grid)
    propagate_IOB_SING(database, tiles_by_grid)
    propagate_IOI_SING(database, tiles_by_grid)
    propagate_IOI_Y9(database, tiles_by_grid)
    alias_HCLKs(database)

    # Save
    xjson.pprint(open(json_out_fn, "w"), database)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Generate tilegrid.json from bitstream deltas")

    parser.add_argument("--verbose", action="store_true", help="")
    parser.add_argument(
        "--json-in",
        default="tiles_basic.json",
        help="Input .json without addresses")
    parser.add_argument(
        "--json-out", default="tilegrid.json", help="Output JSON")
    args = parser.parse_args()

    run(args.json_in, args.json_out, verbose=args.verbose)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import sys, re
import os
from prjxray import util


def run(globaldb, localdb, verbose=False):

    local_db_files = list()
    work_db_files = list()

    # get DB files

    global_entries = {}
    local_entries = {}
    final_entries = {}

    verbose and print("removing %s from %s" % (localdb, globaldb))
    # parse global db
    for line, (tag, bits, mode) in util.parse_db_lines(globaldb):
        global_entries[tag] = bits
    # parse local db
    for line, (tag, bits, mode) in util.parse_db_lines(localdb):
        local_entries[tag] = bits

    for entry in global_entries:
        if entry not in local_entries:
            final_entries[entry] = global_entries[entry]
        else:
            verbose and print("Removing entry %s" % entry)

    util.write_db_lines(globaldb, final_entries)


def main():

    import argparse

    parser = argparse.ArgumentParser(
        description="Remove partial DB from global DB")
    parser.add_argument('--verbose', action='store_true', help='')
    parser.add_argument(
        '--localdb', action='store', help='Path to work database')
    parser.add_argument(
        '--globaldb', action='store', help='Path to global database')

    args = parser.parse_args()

    run(args.globaldb, args.localdb, args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import json
import os

from prjxray.segmaker import Segmaker

BIN = "BIN"
BOOL = "BOOL"


def bitfilter(frame, bit):
    # Filter out interconnect bits.
    if frame not in [28, 29]:
        return False

    return True


def main():
    segmk = Segmaker("design.bits")

    fuz_dir = os.getenv("FUZDIR", None)
    assert fuz_dir
    with open(os.path.join(fuz_dir, "attrs.json"), "r") as attr_file:
        attrs = json.load(attr_file)

    print("Loading tags")
    with open('params.json') as f:
        params = json.load(f)

        site = params['site']

        for param, param_info in attrs.items():
            value = params[param]
            param_type = param_info["type"]
            param_digits = param_info["digits"]
            param_values = param_info["values"]

            if param_type == BIN:
                bitstr = [
                    int(x) for x in "{value:0{digits}b}".format(
                        value=value, digits=param_digits)[::-1]
                ]

                for i in range(param_digits):
                    segmk.add_site_tag(site, "%s[%u]" % (param, i), bitstr[i])
            else:
                assert param_type == BOOL
                segmk.add_site_tag(site, param, value == "TRUE")

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import io
import os
import random
import re
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray import lut_maker
from prjxray import verilog
from prjxray.db import Database


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['BSCAN', 'CAPTURE', 'ICAP', 'USR_ACCESS',
                             'STARTUP', 'FRAME_ECC', 'DCIRESET']:
                if site_name not in 'ICAP_X0Y0':
                    yield site_name, site_type


def write_csv_params(params):
    pinstr = 'tile,site,\n'
    for vals in params:
        pinstr += ','.join(map(str, vals)) + '\n'

    open('params.csv', 'w').write(pinstr)


def generate_params():
    bscan_already_on = False
    icap_already_on = False
    tile_params = []
    for loci, (site, site_type) in enumerate(sorted(gen_sites())):
        p = {}
        if site_type in "ICAP" and not icap_already_on:
            p["ICAP_WIDTH"] = verilog.quote(
                random.choice(["X32", "X8", "X16"]))
        elif site_type in "BSCAN" and not bscan_already_on:
            p["JTAG_CHAIN"] = random.randint(1, 4)
            bscan_already_on = True
        elif site_type in "CAPTURE":
            p["ONESHOT"] = verilog.quote(random.choice(["TRUE", "FALSE"]))
        elif site_type in "STARTUP":
            p["PROG_USR"] = verilog.quote(random.choice(["TRUE", "FALSE"]))
        elif site_type in "FRAME_ECC":
            p["FARSRC"] = verilog.quote(random.choice(["FAR", "EFAR"]))
        elif site_type in [
                "DCIRESET", "USR_ACCESS"
        ]:  #The primitives from these sites have no parameters
            p["ENABLED"] = random.randint(0, 1)
        else:
            continue
        p["LOC"] = verilog.quote(site)
        tile_params.append(
            {
                "site": site,
                "site_type": site_type,
                "module": "mod_{}".format(site_type),
                "params": p
            })
    return tile_params


def generate_netlist(params):
    DUTN = len(params)
    DIN_N = DUTN * 32
    DOUT_N = DUTN * 32

    string_output = io.StringIO()
    any_bscan = False
    any_icap = False
    usr_access_on = False
    capture_on = False
    startup_on = False
    frame_ecc_on = False
    dcireset_on = False
    luts = lut_maker.LutMaker()
    verilog.top_harness(DIN_N, DOUT_N)
    print(
        '''
module roi(input clk, input [%d:0] din, output [%d:0] dout);''' %
        (DIN_N - 1, DOUT_N - 1))
    for loci, param in enumerate(params):
        ports = {
            'din': 'din[{} +: 8]'.format(8 * loci),
            'dout': 'dout[{} +: 8]'.format(8 * loci),
            'clk': 'clk'
        }
        if param["site_type"] in "BSCAN":
            ports = {
                'din':
                '{{din[{} +: 7],{}}}'.format(
                    8 * loci + 1, luts.get_next_output_net()),
                'dout':
                '{{dout[{} +: 7],{}}}'.format(
                    8 * loci + 1, luts.get_next_input_net()),
                'clk':
                'clk'
            }
            any_bscan = True
        elif param["site_type"] in ["ICAP"]:
            any_icap = True
        elif param["site_type"] in ["CAPTURE"]:
            capture_on = True
        elif param["site_type"] in ["STARTUP"]:
            startup_on = True
        elif param["site_type"] in ["FRAME_ECC"]:
            frame_ecc_on = True
        elif param["site_type"] in ["USR_ACCESS", "DCIRESET"]:
            if not param["params"]["ENABLED"]:
                continue
            if param["site_type"] in ["DCIRESET"]:
                dcireset_on = True
            else:
                usr_access_on = True
        else:
            continue
        verilog.instance(
            param["module"],
            "inst_{}".format(param["site"]),
            ports,
            param["params"],
            string_buffer=string_output)

    #Generate LUTs
    for l in luts.create_wires_and_luts():
        print(l)
    print(string_output.getvalue())

    print(
        '''
endmodule

// ---------------------------------------------------------------------''')
    if any_icap:
        print(
            '''
module mod_ICAP (input [7:0] din, output [7:0] dout, input clk);
    parameter ICAP_WIDTH = "X32";
    parameter LOC = "ICAP_X0Y0";

    wire [23:0] icap_out;
    (* KEEP, DONT_TOUCH, LOC=LOC *)
    ICAPE2 #(
    .ICAP_WIDTH(ICAP_WIDTH),
    .SIM_CFG_FILE_NAME("NONE")
    )
    ICAPE2_inst (
    .O({icap_out, dout}),
    .CLK(clk),
    .CSIB(),
    .I({24'd0, din}),
    .RDWRB()
    );
endmodule
''')

    if capture_on:
        print(
            '''
module mod_CAPTURE (input [7:0] din, output [7:0] dout, input clk);
    parameter ONESHOT ="TRUE";
    parameter LOC = "ICAP_X0Y0";
    (* KEEP, DONT_TOUCH, LOC=LOC *)
    CAPTUREE2 #(
    .ONESHOT(ONESHOT) // Specifies the procedure for performing single readback per CAP trigger.
    )
    CAPTUREE2_inst (
    .CAP(1'b0),
    .CLK(clk)
    );
endmodule
''')

    if usr_access_on:
        print(
            '''
module mod_USR_ACCESS (input [7:0] din, output [7:0] dout, input clk);
    parameter ENABLED = 1;
    parameter LOC = "USR_ACCESS_X0Y0";

    wire [23:0] usr_access_wire;

    (* KEEP, DONT_TOUCH, LOC=LOC *)
    USR_ACCESSE2 USR_ACCESSE2_inst (
    .CFGCLK(),
    .DATA({usr_access_wire, dout}),
    .DATAVALID()
    );
endmodule
''')

    if any_bscan:
        print(
            '''
module mod_BSCAN (input [7:0] din, output [7:0] dout, input clk);
    parameter JTAG_CHAIN  = 1;
    parameter LOC = "BSCAN_X0Y0";

    (* KEEP, DONT_TOUCH, LOC=LOC *)
    BSCANE2 #(
    .JTAG_CHAIN(JTAG_CHAIN)
    )
    dut (
    .CAPTURE(),
    .DRCK(),
    .RESET(),
    .RUNTEST(),
    .SEL(),
    .SHIFT(),
    .TCK(),
    .TDI(dout[0]),
    .TMS(),
    .UPDATE(),
    .TDO(din[0])
    );
endmodule
''')

    if startup_on:
        print(
            '''
module mod_STARTUP (input [7:0] din, output [7:0] dout, input clk);
    parameter LOC = "STARTUP_X0Y0";
    parameter PROG_USR = "FALSE";

    (* KEEP, DONT_TOUCH, LOC=LOC *)
    STARTUPE2 #(
    .PROG_USR(PROG_USR), // Activate program event security feature. Requires encrypted bitstreams.
    .SIM_CCLK_FREQ(0.0) // Set the Configuration Clock Frequency(ns) for simulation.
    )
    STARTUPE2_inst (
    .CFGCLK(),
    .CFGMCLK(),
    .EOS(),
    .PREQ(dout[0]),
    .CLK(clk),
    .GSR(),
    .GTS(),
    .KEYCLEARB(),
    .PACK(),
    .USRCCLKO(),
    .USRCCLKTS(),
    .USRDONEO(),
    .USRDONETS()
    );
endmodule
''')

    if frame_ecc_on:
        print(
            '''
module mod_FRAME_ECC (input [7:0] din, output [7:0] dout, input clk);
    parameter LOC = "FRAME_ECC_X0Y0";
    parameter FARSRC = "EFAR";

    wire [25:0] far_wire;
    assign dout[7:0] = far_wire[7:0];
    (* KEEP, DONT_TOUCH, LOC=LOC *)
    FRAME_ECCE2 #(
    .FARSRC(FARSRC),
    .FRAME_RBT_IN_FILENAME("NONE")
    )
    FRAME_ECCE2_inst (
    .CRCERROR(),
    .ECCERROR(),
    .ECCERRORSINGLE(),
    .FAR(far_wire),
    .SYNBIT(),
    .SYNDROME(),
    .SYNDROMEVALID(),
    .SYNWORD()
    );
endmodule
''')

    if dcireset_on:
        print(
            '''
module mod_DCIRESET (input [7:0] din, output [7:0] dout, input clk);
    parameter LOC = "FRAME_ECC_X0Y0";
    parameter ENABLED = 1;

    (* KEEP, DONT_TOUCH, LOC=LOC *)
    DCIRESET DCIRESET_inst (
    .LOCKED(dout[0]),
    .RST(dout[1])
);
endmodule
''')


def run():
    params = generate_params()
    generate_netlist(params)
    with open('params.jl', 'w') as f:
        json.dump(params, f, indent=2)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
Take raw .bits files and decode them to higher level functionality

A segment as currently used is defined as a tile type + a memory region
Ex: BRAM_L_X6Y100:CLB_IO_CLK
'''

import sys, os, json, re
import copy
from prjxray import bitstream
from prjxray import db as prjxraydb
from prjxray.util import OpenSafeFile, parse_tagbit, db_root_arg, part_arg


class NoDB(Exception):
    pass


# cache
segbitsdb = dict()


# int and sites are loaded together so that bit coverage can be checked together
# however, as currently written, each segment is essentially printed twice
def process_db(db, tile_type, process, verbose):
    ttdb = db.get_tile_type(tile_type)

    fns = [ttdb.tile_dbs.segbits, ttdb.tile_dbs.ppips]
    verbose and print("process_db(%s): %s" % (tile_type, fns))
    for fn in fns:
        if fn:
            with OpenSafeFile(fn, "r") as f:
                for line in f:
                    process(line)


def get_database(db, tile_type, bit_only=False, verbose=False):
    tags = list()

    if tile_type in segbitsdb:
        return segbitsdb[tile_type]

    def process(l):
        # l like: CLBLL_L.SLICEL_X0.AMUX.CY !30_07 !30_11 30_06 30_08
        # Parse tags to do math when multiple tiles share an address space
        parts = l.split()
        name = parts[0]

        if parts[1] == 'always' or parts[1] == 'hint' or parts[1] == 'default':
            if bit_only:
                return
            tagbits = []
        else:
            tagbits = [parse_tagbit(x) for x in parts[1:]]

        tags.append(list([name] + tagbits))

    process_db(db, tile_type, process, verbose=verbose)

    if len(tags) == 0:
        raise NoDB(tile_type)

    segbitsdb[tile_type] = tags
    return tags


def mk_segbits(seginfo, bitdata):
    '''
    Given a tile memory region (seginfo), return list of bits in that region

    seginfo: mk_segments()s object supplying address range
    bitdata: all bits in the entire bitstream
    '''

    segbits = set()

    block = seginfo["block"]
    baseaddr = int(block["baseaddr"], 0)
    frames = block["frames"]
    word_offset = block["offset"]
    words = block["words"]

    for frame in range(baseaddr, baseaddr + frames):
        if frame not in bitdata:
            continue
        for wordidx in range(word_offset, word_offset + words):
            if wordidx not in bitdata[frame]:
                continue
            for bitidx in bitdata[frame][wordidx]:
                frame_addr = frame - baseaddr
                bit_addr = 32 * (wordidx - word_offset) + bitidx
                #segbits.add( "%02d_%02d" % (frame_addr, word_addr))
                segbits.add((frame_addr, bit_addr))

    return segbits


def gen_tilegrid_masks(tiles):
    """yield (addr_min, addr_max + 1, word_min, word_max + 1)"""
    for tilek, tilev in tiles.items():
        for block_type, blockj in tilev["bits"].items():
            baseaddr = int(blockj["baseaddr"], 0)
            frames = blockj["frames"]
            offset = blockj["offset"]
            words = blockj["words"]
            yield (baseaddr, baseaddr + frames, offset, offset + words)


def print_unknown_bits(tiles, bitdata):
    '''
    Print bits not covered by known tiles

    tiles: tilegrid json
    bitdata[addr][word] = set of bit indices (0 to 31)
    '''
    # Start with an open set and remove elements as we find them
    tocheck = copy.deepcopy(bitdata)

    for addr_min, addr_max_p1, word_min, word_max_p1 in gen_tilegrid_masks(
            tiles):
        for addr in range(addr_min, addr_max_p1):
            if addr not in tocheck:
                continue
            for word in range(word_min, word_max_p1):
                if word not in tocheck[addr]:
                    continue
                del tocheck[addr][word]

    # print uncovered locations
    print('Non-database bits:')
    for frame in sorted(tocheck.keys()):
        for wordidx in sorted(tocheck[frame].keys()):
            for bitidx in sorted(tocheck[frame][wordidx]):
                print("bit_%08x_%03d_%02d" % (frame, wordidx, bitidx))


def tagmatch(entry, segbits):
    '''Does tag appear in segbits?'''

    # Entry like "CLBLL_L.SLICEL_X0.AMUX.CY !30_07 !30_11 30_06 30_08".split()
    for bit in entry[1:]:
        isset, bitaddr = bit

        # Reject if bit polarity is incorrect
        if bitaddr not in segbits if isset else bitaddr in segbits:
            return False
    return True


def tag_matched(entry, segbits):
    for bit in entry[1:]:
        isset, bitaddr = bit
        if isset:
            segbits.remove(bitaddr)


# tile types that failed to decode
decode_warnings = set()


def seg_decode(db, seginfo, segbits, segments, bit_only=False, verbose=False):
    '''
    Remove matched tags from segbits
    Returns a list of all matched tags
    '''

    segtags = set()

    # Valid addresses for refereced tiles
    ref_block = seginfo["block"]
    # ref_frame_as = (int(ref_block["baseaddr"], 0), int(ref_block["baseaddr"], 0) + ref_block["frames"] - 1)
    ref_frame_as = (0, ref_block["frames"] - 1)
    ref_bit_as = (
        32 * ref_block["offset"],
        32 * (ref_block["offset"] + ref_block["words"]) - 1)

    def process(cmp_seginfo, ref_tile_name):
        tile_type = cmp_seginfo["tile"]["type"]

        # already failed?
        if tile_type in decode_warnings:
            return

        try:
            entries = get_database(
                db, tile_type, bit_only=bit_only, verbose=verbose)
        except NoDB:
            verbose and print("WARNING: failed to load DB for %s" % tile_type)
            assert tile_type != 'BRAM_L'
            decode_warnings.add(tile_type)
            return

        cmp_block = cmp_seginfo["block"]
        ref_frame_delta = int(cmp_block["baseaddr"], 0) - int(
            ref_block["baseaddr"], 0)
        ref_bit_delta = 32 * (cmp_block["offset"] - ref_block["offset"])

        def adjust_entry_addr(entry):
            '''Return bits that apply to this tile at the correct address'''
            tagname = entry[0]
            bits = entry[1:]
            if len(bits) == 0:
                return None
            ret = [tagname]

            def adjust_entry():
                '''Adjust entry in another tile address space to be in our reference tile address space'''
                for isset, old_bitaddr, in bits:
                    old_frame_addr, old_bit_addr = old_bitaddr
                    new_frame_addr = old_frame_addr + ref_frame_delta
                    if not (ref_frame_as[0] <= new_frame_addr <=
                            ref_frame_as[1]):
                        verbose and print(
                            "out frame range: %d <= %d <= %d" %
                            (ref_frame_as[0], new_frame_addr, ref_frame_as[1]))
                        return False

                    new_bit_addr = old_bit_addr + ref_bit_delta
                    # Verify in range of original tile
                    # This can happen if a smaller tile references a larger tile
                    if not (ref_bit_as[0] <= new_bit_addr <= ref_bit_as[1]):
                        verbose and print(
                            "out bit range: %d <= %d <= %d" %
                            (ref_bit_as[0], new_bit_addr, ref_bit_as[1]))
                        return False
                    ret.append((isset, (new_frame_addr, new_bit_addr)))
                    verbose and print(
                        "ent %02d_%02d => %02d_%02d" % (
                            old_frame_addr, old_bit_addr, new_frame_addr,
                            new_bit_addr))
                return True

            if not adjust_entry():
                return None
            return ret

        for entry in entries:
            if ref_tile_name:
                entry = adjust_entry_addr(entry)
                if entry is None:
                    continue
                verbose and print('adjusted entry', entry)

            if not tagmatch(entry, segbits):
                continue
            tag_matched(entry, segbits)
            tagname = entry[0]
            # Prefix matches not from this tile
            if ref_tile_name:
                segtags.add('%s:%s' % (ref_tile_name, tagname))
            else:
                segtags.add(tagname)

    # Reference tile
    process(seginfo, None)
    # Tiles that share our address space
    for (ref_tile_name, cmp_block_name) in seginfo['segtiles']:
        process(
            segments[mksegment(ref_tile_name, cmp_block_name)], ref_tile_name)

    return segtags


def print_seg(
        segname, seginfo, nbits, segbits, segtags, decode_emit, verbose=False):
    '''Print segment like used by segmaker/segmatch'''

    print("seg %s" % (segname, ))
    if verbose:
        print("Bits: %s" % nbits)
        print(
            "Address: %s, +%s" %
            (seginfo["block"]["baseaddr"], seginfo["block"]["frames"]))
        print(
            "Words: %s, +%s" %
            (seginfo["block"]["offset"], seginfo["block"]["words"]))

    # Bits that weren't decoded
    for bit in sorted(segbits):
        print("bit %02d_%02d" % bit)

    if decode_emit:
        for tag in sorted(segtags):
            print("tag %s" % tag)


def handle_segment(
        db,
        segname,
        bitdata,
        decode_emit,
        decode_omit,
        omit_empty_segs,
        segments,
        bit_only=False,
        verbose=False):

    seginfo = segments[segname]

    segbits = mk_segbits(seginfo, bitdata)
    nbits = len(segbits)

    if decode_emit or decode_omit:
        segtags = seg_decode(
            db, seginfo, segbits, segments, bit_only=bit_only, verbose=verbose)
    else:
        segtags = set()

    # Found something to print?
    keep = not omit_empty_segs or len(segbits) > 0 or (
        len(segtags) > 0 and not decode_omit)
    if not keep:
        return

    print()
    print_seg(
        segname,
        seginfo,
        nbits,
        segbits,
        segtags,
        decode_emit,
        verbose=verbose)


def overlap(a, b):
    return a[0] <= b[0] <= a[1] or b[0] <= a[0] <= b[1]


def mk_segtiles(tiles):
    '''
    Return a dictionary of tile_name:tiles
    Where tiles is a list of tiles that are in our address space

    Assumption: tiles in the same minor address region have the same base address and number frames

    As DB is written, not all have the same number of frames
    Ex: CLBLM_R_X7Y108 36 frames, INT_R_X7Y108 28 frames
    We could check for this, but don't think its worth the effort
    Maybe this should be corrected in the DB?
    '''

    segtiles = {}

    # Group by base address
    baseaddrs = {}
    for tile_name, tile in tiles.items():
        for block_name, block in tile['bits'].items():
            baseaddrs.setdefault(block["baseaddr"], []).append(
                (block["offset"], tile_name, block, block_name))

    for baseaddr, values in baseaddrs.items():
        '''
        There are only 256 addresses per minor address
        Just do a set brute force search for now?
        Maybe too slow with the number of tiles

        Around 50 IP blocks max per minor address
        '''

        # Sort by block offset
        values = sorted(values)

        for refi, (_ref_block_offset, ref_tile_name, ref_block,
                   ref_block_name) in enumerate(values):
            seglets = segtiles.setdefault(ref_tile_name, [])
            ref_as = (
                ref_block["offset"],
                ref_block["offset"] + ref_block["words"] - 1)

            for cmpi in range(refi + 1, len(values)):
                (_cmp_block_offset, cmp_tile_name, cmp_block,
                 cmp_block_name) = values[cmpi]
                cmp_as = (
                    cmp_block["offset"],
                    cmp_block["offset"] + cmp_block["words"] - 1)

                if overlap(ref_as, cmp_as):
                    seglets.append((cmp_tile_name, cmp_block_name))
                # sorting => first non-intersection means no future will intersect
                else:
                    break

    return segtiles


def mk_segments(tiles):
    segments = {}
    segtiles = mk_segtiles(tiles)

    for tile_name, tile in tiles.items():
        for block_name, block in tile['bits'].items():
            segname = mksegment(tile_name, block_name)
            segments[segname] = {
                'tile': tile,
                'tile_name': tile_name,
                'block': block,
                'block_name': block_name,
                'segtiles': segtiles[tile_name],
            }
    return segments


def mksegment(tile_name, block_name):
    '''Create a segment name'''
    return '%s:%s' % (tile_name, block_name)


def tile_segnames(tiles):
    '''Create a list of all (tile_name, block_name) from input tiles'''
    ret = []
    for tile_name, tile in tiles.items():
        if 'bits' not in tile:
            continue

        for block_name in tile['bits'].keys():
            ret.append(mksegment(tile_name, block_name))
    return ret


def load_tiles(db_root, part):
    # TODO: Migrate to new tilegrid format via library.
    with OpenSafeFile("%s/%s/tilegrid.json" % (db_root, part), "r") as f:
        tiles = json.load(f)
    return tiles


def run(
        db_root,
        part,
        bits_file,
        segnames,
        omit_empty_segs=False,
        flag_unknown_bits=False,
        flag_decode_emit=False,
        flag_decode_omit=False,
        bit_only=False,
        verbose=False):
    db = prjxraydb.Database(db_root, part)
    tiles = load_tiles(db_root, part)
    segments = mk_segments(tiles)
    with OpenSafeFile(bits_file) as f:
        bitdata = bitstream.load_bitdata2(f)

    if flag_unknown_bits:
        print_unknown_bits(tiles, bitdata)
        print("")

    # Default: print all
    if segnames:
        for i, segname in enumerate(segnames):
            # Default to common tile config area if tile given without explicit block
            if ':' not in segname:
                segnames[i] = mksegment(segname, 'CLB_IO_CLK')
    else:
        segnames = sorted(tile_segnames(tiles))
    print('Segments: %u' % len(segnames))

    # XXX: previously this was sorted by address, not name
    # revisit?
    for segname in segnames:
        handle_segment(
            db,
            segname,
            bitdata,
            flag_decode_emit,
            flag_decode_omit,
            omit_empty_segs,
            segments,
            bit_only=bit_only,
            verbose=verbose)


def main():
    import argparse

    parser = argparse.ArgumentParser(
        description="Decode bits within a tile's address space")

    db_root_arg(parser)
    part_arg(parser)
    parser.add_argument('--verbose', action='store_true', help='')
    parser.add_argument(
        '-z',
        action='store_true',
        help="do not print a 'seg' header for empty segments")
    parser.add_argument(
        '-b', action='store_true', help='print bits outside of known segments')
    parser.add_argument(
        '-d',
        action='store_true',
        help='decode known segment bits and write them as tags')
    parser.add_argument(
        '-D',
        action='store_true',
        help='decode known segment bits and omit them in the output')
    parser.add_argument(
        '--bit-only',
        action='store_true',
        help='only decode real bitstream directives')
    parser.add_argument('bits_file', help='')
    parser.add_argument(
        'segnames', nargs='*', help='List of tile or tile:block to print')
    args = parser.parse_args()

    run(
        args.db_root,
        args.part,
        args.bits_file,
        args.segnames,
        args.z,
        args.b,
        args.d,
        args.D,
        bit_only=args.bit_only,
        verbose=args.verbose)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import os, random
random.seed(int(os.getenv("SEED"), 16))

import re
import json

from prjxray import util
from prjxray.db import Database

# =============================================================================


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()

    tile_list = []
    for tile_name in sorted(grid.tiles()):
        if "IOB33" not in tile_name or "SING" in tile_name:
            continue
        tile_list.append(tile_name)

    get_xy = util.create_xy_fun('[LR]IOB33_')
    tile_list.sort(key=get_xy)

    for iob_tile_name in tile_list:
        iob_gridinfo = grid.gridinfo_at_loc(
            grid.loc_of_tilename(iob_tile_name))

        # Find IOI tile adjacent to IOB
        for suffix in ["IOI3", "IOI3_TBYTESRC", "IOI3_TBYTETERM"]:
            try:
                ioi_tile_name = iob_tile_name.replace("IOB33", suffix)
                ioi_gridinfo = grid.gridinfo_at_loc(
                    grid.loc_of_tilename(ioi_tile_name))
                break
            except KeyError:
                pass

        iob33s = [k for k, v in iob_gridinfo.sites.items() if v == "IOB33S"][0]
        iob33m = [k for k, v in iob_gridinfo.sites.items() if v == "IOB33M"][0]
        idelay_s = iob33s.replace("IOB", "IDELAY")
        idelay_m = iob33m.replace("IOB", "IDELAY")

        yield iob33m, idelay_m, iob33s, idelay_s


def run():

    # Get all [LR]IOI3 tiles
    tiles = list(gen_sites())

    # Header
    print("// Tile count: %d" % len(tiles))
    print("// Seed: '%s'" % os.getenv("SEED"))

    ninputs = 0
    di_idx = []
    for i, sites in enumerate(tiles):
        if random.randint(0, 1):
            di_idx.append(ninputs)
            ninputs += 1
        else:
            di_idx.append(None)

    print(
        '''
module top (
  (* CLOCK_BUFFER_TYPE = "NONE" *)
  input  wire clk,
  input  wire [{N}:0] di
);

wire clk_buf = clk;

wire [{N}:0] di_buf;
    '''.format(N=ninputs - 1))

    # LOCes IOBs
    data = []
    for i, (sites, ibuf_idx) in enumerate(zip(tiles, di_idx)):

        if random.randint(0, 1):
            iob_i = sites[0]
            iob_o = sites[2]
            idelay = sites[1]
            other_idelay = sites[3]
        else:
            iob_i = sites[2]
            iob_o = sites[0]
            idelay = sites[3]
            other_idelay = sites[1]

        use_ibuf = ibuf_idx is not None

        DELAY_SRC = random.choice(["IDATAIN", "DATAIN"])
        if not use_ibuf:
            DELAY_SRC = 'DATAIN'

        params = {
            "LOC":
            "\"" + idelay + "\"",
            "IDELAY_TYPE":
            "\"" + random.choice(
                ["FIXED", "VARIABLE", "VAR_LOAD", "VAR_LOAD_PIPE"]) + "\"",
            "IDELAY_VALUE":
            random.randint(0, 31),
            "DELAY_SRC":
            "\"" + DELAY_SRC + "\"",
            "HIGH_PERFORMANCE_MODE":
            "\"" + random.choice(["TRUE", "FALSE"]) + "\"",
            "CINVCTRL_SEL":
            "\"" + random.choice(["TRUE", "FALSE"]) + "\"",
            "PIPE_SEL":
            "\"" + random.choice(["TRUE", "FALSE"]) + "\"",
            "IS_C_INVERTED":
            random.randint(0, 1),
            "IS_DATAIN_INVERTED":
            random.randint(0, 1),
            "IS_IDATAIN_INVERTED":
            random.randint(0, 1),
        }

        if params["IDELAY_TYPE"] != "\"VAR_LOAD_PIPE\"":
            params["PIPE_SEL"] = "\"FALSE\""

        # The datasheet says that for these two modes the delay is set to 0
        if params["IDELAY_TYPE"] == "\"VAR_LOAD\"":
            params["IDELAY_VALUE"] = 0
        if params["IDELAY_TYPE"] == "\"VAR_LOAD_PIPE\"":
            params["IDELAY_VALUE"] = 0

        if params["IDELAY_TYPE"] == "\"FIXED\"":
            params["IS_C_INVERTED"] = 0

        param_str = ",".join(".%s(%s)" % (k, v) for k, v in params.items())

        if use_ibuf:
            print('')
            print('(* LOC="%s", KEEP, DONT_TOUCH *)' % iob_i)
            print(
                'IBUF ibuf_%03d (.I(di[%3d]), .O(di_buf[%3d]));' %
                (ibuf_idx, ibuf_idx, ibuf_idx))
            print(
                'mod #(%s) mod_%03d (.clk(clk_buf), .I(di_buf[%3d]));' %
                (param_str, i, ibuf_idx))
        else:
            print('mod #(%s) mod_%03d (.clk(clk_buf), .I());' % (param_str, i))

        params['IBUF_IN_USE'] = use_ibuf
        params["IDELAY_IN_USE"] = idelay
        params["IDELAY_NOT_IN_USE"] = other_idelay

        data.append(params)

    # Store params
    with open("params.json", "w") as fp:
        json.dump(data, fp, sort_keys=True, indent=1)

    print(
        '''
// IDELAYCTRL
(* KEEP, DONT_TOUCH *)
IDELAYCTRL idelayctrl();

endmodule

(* KEEP, DONT_TOUCH *)
module mod(
  input  wire clk,
  input  wire I
);

parameter LOC = "";
parameter IDELAY_TYPE = "FIXED";
parameter IDELAY_VALUE = 0;
parameter DELAY_SRC = "IDATAIN";
parameter HIGH_PERFORMANCE_MODE = "TRUE";
parameter SIGNAL_PATTERN = "DATA";
parameter CINVCTRL_SEL = "FALSE";
parameter PIPE_SEL = "FALSE";
parameter IS_C_INVERTED = 0;
parameter IS_DATAIN_INVERTED = 0;
parameter IS_IDATAIN_INVERTED = 0;

wire x;
wire lut;

(* KEEP, DONT_TOUCH *)
LUT2 l( .O(lut) );

// IDELAY
(* LOC=LOC, KEEP, DONT_TOUCH *)
IDELAYE2 #(
  .IDELAY_TYPE(IDELAY_TYPE),
  .IDELAY_VALUE(IDELAY_VALUE),
  .DELAY_SRC(DELAY_SRC),
  .HIGH_PERFORMANCE_MODE(HIGH_PERFORMANCE_MODE),
  .SIGNAL_PATTERN(SIGNAL_PATTERN),
  .CINVCTRL_SEL(CINVCTRL_SEL),
  .PIPE_SEL(PIPE_SEL),
  .IS_C_INVERTED(IS_C_INVERTED),
  .IS_DATAIN_INVERTED(IS_DATAIN_INVERTED),
  .IS_IDATAIN_INVERTED(IS_IDATAIN_INVERTED)
)
idelay
(
  .C(clk),
  .REGRST(),
  .LD(),
  .CE(),
  .INC(),
  .CINVCTRL(),
  .CNTVALUEIN(),
  .IDATAIN(I),
  .DATAIN(lut),
  .LDPIPEEN(),
  .DATAOUT(x),
  .CNTVALUEOUT()
);

endmodule
    ''')


run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" """
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database


class PipList(object):
    def __init__(self):
        self.piplist = {}
        self.ppiplist = {}

    def get_pip_and_ppip_list_for_tile_type(self, tile_type):

        # Load PIP list for the tile type if not already loaded
        if tile_type not in self.piplist:
            self.piplist[tile_type] = []

            fname = os.path.join(
                os.getenv('FUZDIR'), '..', 'piplist', 'build', 'cmt_top_lower',
                tile_type.lower() + '.txt')

            with open(fname, "r") as f:
                for l in f:
                    pip, is_directional = l.strip().split(' ')
                    tile, dst, src = pip.split('.')
                    if tile_type == tile:
                        self.piplist[tile_type].append(
                            (src, dst, bool(int(is_directional))))

        # Load PPIP list for the tile type if not already loaded
        if tile_type not in self.ppiplist:
            self.ppiplist[tile_type] = []

            fname = os.path.join(
                os.getenv('FUZDIR'), '..', '071-ppips', 'build',
                'ppips_' + tile_type.lower() + '.db')

            with open(fname, "r") as f:
                for l in f:
                    pip_data, pip_type = l.strip().split()

                    if pip_type != 'always':
                        continue

                    tile, dst, src = pip_data.strip().split('.')
                    if tile_type == tile:
                        self.ppiplist[tile_type].append((src, dst, True))

        return self.piplist[tile_type], self.ppiplist[tile_type]


def find_phasers_for_mmcm(grid, loc):
    gridinfo = grid.gridinfo_at_loc((loc[0], loc[1] - 9))

    phasers = {
        'IN': [],
        'OUT': [],
    }

    for site_name, site_type in gridinfo.sites.items():
        if site_type == 'PHASER_IN_PHY':
            phasers['IN'].append(site_name)
        elif site_type == 'PHASER_OUT_PHY':
            phasers['OUT'].append(site_name)

    assert len(phasers['IN']) > 0
    assert len(phasers['OUT']) > 0

    phasers['IN'].sort()
    phasers['OUT'].sort()

    return phasers


def find_hclk_ref_wires_for_mmcm(grid, loc):
    tilename = grid.tilename_at_loc((loc[0], loc[1] - 17))
    gridinfo = grid.gridinfo_at_tilename(tilename)

    assert gridinfo.tile_type in ['HCLK_CMT_L', 'HCLK_CMT']

    # HCLK_CMT_MUX_OUT_FREQ_REF[0-3]
    wires = []
    for idx in range(4):
        wires.append('{}/HCLK_CMT_MUX_OUT_FREQ_REF{}'.format(tilename, idx))

    return wires


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)

        for site_name, site_type in gridinfo.sites.items():
            if site_type in ['MMCME2_ADV']:
                phasers = find_phasers_for_mmcm(grid, loc)
                hclk_wires = find_hclk_ref_wires_for_mmcm(grid, loc)
                yield tile_name, site_name, phasers, hclk_wires


def get_random_route_from_site_pin(
        pip_list, tile_name, site_pin, direction, occupied_wires, hclk_wires):

    # A map of MMCM site pins to wires they are connected to.
    pin_to_wire = {
        "CMT_TOP_L_LOWER_B": {
            "CLKIN1": "CMT_LR_LOWER_B_MMCM_CLKIN1",
            "CLKIN2": "CMT_LR_LOWER_B_MMCM_CLKIN2",
            "CLKFBIN": "CMT_LR_LOWER_B_MMCM_CLKFBIN",
            "CLKOUT0": "CMT_LR_LOWER_B_MMCM_CLKOUT0",
            "CLKOUT1": "CMT_LR_LOWER_B_MMCM_CLKOUT1",
            "CLKOUT2": "CMT_LR_LOWER_B_MMCM_CLKOUT2",
            "CLKOUT3": "CMT_LR_LOWER_B_MMCM_CLKOUT3",
            "CLKOUT4": "CMT_LR_LOWER_B_MMCM_CLKOUT4",
            "CLKOUT5": "CMT_LR_LOWER_B_MMCM_CLKOUT5",
            "CLKOUT6": "CMT_LR_LOWER_B_MMCM_CLKOUT6",
        },
        "CMT_TOP_R_LOWER_B": {
            "CLKIN1": "CMT_LR_LOWER_B_MMCM_CLKIN1",
            "CLKIN2": "CMT_LR_LOWER_B_MMCM_CLKIN2",
            "CLKFBIN": "CMT_LR_LOWER_B_MMCM_CLKFBIN",
            "CLKOUT0": "CMT_LR_LOWER_B_MMCM_CLKOUT0",
            "CLKOUT1": "CMT_LR_LOWER_B_MMCM_CLKOUT1",
            "CLKOUT2": "CMT_LR_LOWER_B_MMCM_CLKOUT2",
            "CLKOUT3": "CMT_LR_LOWER_B_MMCM_CLKOUT3",
            "CLKOUT4": "CMT_LR_LOWER_B_MMCM_CLKOUT4",
            "CLKOUT5": "CMT_LR_LOWER_B_MMCM_CLKOUT5",
            "CLKOUT6": "CMT_LR_LOWER_B_MMCM_CLKOUT6",
        },
    }

    # Get tile type
    tile_type = tile_name.rsplit("_", maxsplit=1)[0]

    # Get all PIPs (PIPs + PPIPs)
    pips, ppips = pip_list.get_pip_and_ppip_list_for_tile_type(tile_type)
    all_pips = pips + ppips

    # The first wire
    wire = pin_to_wire[tile_type][site_pin]

    if site_pin in ["CLKIN1", "CLKIN2", "CLKFBIN"] and random.random() < .2:
        return [random.choice(hclk_wires), wire]

    # Walk randomly.
    route = []
    while True:
        route.append(wire)

        wires = []

        for src, dst, is_directional in all_pips:
            if direction == "down" and src == wire:
                next_wire = dst
            elif direction == "down" and dst == wire and not is_directional:
                next_wire = src
            elif direction == "up" and dst == wire:
                next_wire = src
            elif direction == "up" and src == wire and not is_directional:
                next_wire = dst
            else:
                continue

            if next_wire not in occupied_wires:
                wires.append(next_wire)

        if len(wires) == 0:
            break

        wire = random.choice(wires)
        occupied_wires.add(wire)

    # For "up" direction reverse the route.
    if direction == "down":
        return route
    if direction == "up":
        return route[::-1]


def main():

    # 8 inputs per clock region
    # 5 clock regions for device
    max_clk_inputs = 8 * 5
    clkin_idx = 0

    print(
        '''
module top(
  input wire [{nclkin}:0] clkin
);

    (* KEEP, DONT_TOUCH *)
    LUT6 dummy();
    '''.format(nclkin=max_clk_inputs - 1))

    pip_list = PipList()
    bufg_count = 0

    design_file = open('design.txt', 'w')
    routes_file = open('routes.txt', 'w')

    count = 0
    mmcms = sorted(gen_sites(), key=lambda x: x[0])
    random.shuffle(mmcms)
    for tile, site, phasers, hclk_wires in mmcms:
        in_use = random.randint(0, 2) > 0
        count += 1

        design_file.write("{},{},{}\n".format(tile, site, int(in_use)))

        if not in_use:
            continue

        # Generate random routes to/from some pins
        routes = {}
        endpoints = {}

        pins = [
            ('CLKIN1', 'up'),
            ('CLKIN2', 'up'),
            ('CLKFBIN', 'up'),

            # Sometimes manually randomized route for CLKOUTx conflicts with
            # the verilog design.
            #('CLKOUT0', 'down'),
            #('CLKOUT1', 'down'),
            #('CLKOUT2', 'down'),
            #('CLKOUT3', 'down'),
            #('CLKOUT4', 'down'),
            #('CLKOUT5', 'down'),
            #('CLKOUT6', 'down'),
        ]

        occupied_wires = set()
        for pin, dir in pins:

            route = get_random_route_from_site_pin(
                pip_list, tile, pin, dir, occupied_wires, hclk_wires)
            if route is None:
                endpoints[pin] = ""
                continue

            routes[pin] = (
                route,
                dir,
            )
            endpoints[pin] = route[-1] if dir == 'down' else route[0]

        internal_feedback = endpoints['CLKFBIN'].endswith('CLKFBOUT')

        # Store them in a random order so the TCL script will try to route
        # them also in the random order.
        lines = []
        pins = sorted(routes.keys())
        random.shuffle(pins)
        for pin in pins:
            (route, dir) = routes[pin]
            route_str = " ".join(route)
            lines.append(
                '{} {} {} {} {}\n'.format(tile, site, pin, dir, route_str))

        random.shuffle(lines)
        routes_file.writelines(lines)

        clkfbin_src = random.choice(('BUFH', 'logic'))

        if internal_feedback:
            COMPENSATION = "INTERNAL"
        else:
            if clkfbin_src == 'logic':
                COMPENSATION = 'EXTERNAL'
            else:
                COMPENSATION = "ZHOLD"

        print(
            """
    wire clkfbin_{site};
    wire clkin1_{site};
    wire clkin2_{site};
    wire clkfbout_mult_{site};
    wire clkout0_{site};
    wire clkout1_{site};
    wire clkout2_{site};
    wire clkout3_{site};
    wire clkout4_{site};
    wire clkout5_{site};
    (* KEEP, DONT_TOUCH, LOC = "{site}" *)
    MMCME2_ADV #(
        .COMPENSATION("{COMPENSATION}")
        ) mmcm_{site} (
            .CLKFBOUT(clkfbout_mult_{site}),
            .CLKOUT0(clkout0_{site}),
            .CLKOUT1(clkout1_{site}),
            .CLKOUT2(clkout2_{site}),
            .CLKOUT3(clkout3_{site}),
            .CLKOUT4(clkout4_{site}),
            .CLKOUT5(clkout5_{site}),
            .DRDY(),
            .LOCKED(),
            .DO(),
            .CLKFBIN(clkfbin_{site}),
            .CLKIN1(clkin1_{site}),
            .CLKIN2(clkin2_{site}),
            .CLKINSEL(),
            .DCLK(),
            .DEN(),
            .DWE(),
            .PWRDWN(),
            .RST(),
            .DI(),
            .DADDR());
            """.format(site=site, COMPENSATION=COMPENSATION))

        for clkout in range(4, 6):
            # CLKOUT4 and CLKOUT5 can only drive one signal type
            if random.randint(0, 1) and bufg_count < 16:
                bufg_count += 1
                print(
                    """
            (* KEEP, DONT_TOUCH *)
            BUFG (
                .I(clkout{idx}_{site})
            );""".format(idx=clkout, site=site))

        any_phaser = False

        for clkout in range(4):
            # CLKOUT0-CLKOUT3 can drive:
            #  - Global drivers (e.g. BUFG)
            #  - PHASER_[IN|OUT]_[CA|DB]_FREQREFCLK via BB_[0-3]
            drive_bufg = random.randint(0, 1) and bufg_count < 16
            drive_phaser = 0  #random.randint(0, 1)

            if drive_bufg:
                bufg_count += 1
                print(
                    """
            (* KEEP, DONT_TOUCH *)
            BUFG (
                .I(clkout{idx}_{site})
            );""".format(idx=clkout, site=site))

            if drive_phaser and not any_phaser and False:
                any_phaser = True
                print(
                    """
            (* KEEP, DONT_TOUCH, LOC="{phaser_loc}" *)
            PHASER_OUT phaser_{site}(
                .FREQREFCLK(clkout{idx}_{site})
            );""".format(idx=clkout, site=site, phaser_loc=phasers['OUT'][0]))

        if internal_feedback:
            print(
                """
                assign clkfbin_{site} = clkfbout_mult_{site};
                """.format(site=site))
        else:
            if clkfbin_src == 'BUFH':
                print(
                    """
                (* KEEP, DONT_TOUCH *)
                BUFH (
                    .I(clkfbout_mult_{site}),
                    .O(clkfbin_{site})
                );""".format(site=site))
            elif clkfbin_src == 'logic':
                print(
                    """
                (* KEEP, DONT_TOUCH *)
                LUT6 # (.INIT(64'h5555555555555555))
                clkfbin_logic_{site} (
                    .I0(clkfbout_mult_{site}),
                    .O(clkfbin_{site})
                );
                """.format(site=site))
            else:
                assert False, clkfbin_src

        for clkin in range(2):
            clkin_src = random.choice((
                'BUFH',
                'BUFR',
                'logic',
            ))

            if clkin_src == 'BUFH':
                print(
                    """
                (* KEEP, DONT_TOUCH *)
                BUFH (
                    .O(clkin{idx}_{site}),
                    .I(clkin{idx2})
                );""".format(idx=clkin + 1, idx2=clkin_idx, site=site))
            elif clkin_src == 'BUFR':
                print(
                    """
                (* KEEP, DONT_TOUCH *)
                BUFR (
                    .O(clkin{idx}_{site}),
                    .I(clkin{idx2})
                );""".format(idx=clkin + 1, idx2=clkin_idx, site=site))
            elif clkin_src == 'logic':
                print(
                    """
                (* KEEP, DONT_TOUCH *)
                LUT6 # (.INIT(64'h5555555555555555))
                clkin{idx}_logic_{site} (
                    .I0(clkin{idx2}),
                    .O(clkin{idx}_{site})
                );
                """.format(idx=clkin + 1, idx2=clkin_idx, site=site))
            else:
                assert False, (clkin, clkin_src)

            clkin_idx += 1

    print("endmodule")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json

from prjxray.segmaker import Segmaker
from prjxray import util
from prjxray import verilog

iface_types = [
    "NETWORKING", "OVERSAMPLE", "MEMORY", "MEMORY_DDR3", "MEMORY_QDR"
]

data_rates = ["SDR", "DDR"]

data_widths = {
    "SDR": [2, 3, 4, 5, 6, 7, 8],
    "DDR": [4, 6, 8, 10, 14],
}


def run():

    segmk = Segmaker("design.bits")

    # Load tags
    with open("params.json", "r") as fp:
        data = json.load(fp)

    loc_to_tile_site_map = {}

    # Output tags
    for param_list in data:
        for params in param_list:
            loc = verilog.unquote(params["SITE_LOC"])

            get_xy = util.create_xy_fun('IOB_')
            x, y = get_xy(loc.replace("ILOGIC", "IOB"))

            loc_to_tile_site_map[loc] = params["TILE_NAME"] + ".IOB_Y%d" % (
                y % 2)

            # Site not used at all
            if not params["IS_USED"]:

                segmk.add_site_tag(loc, "ISERDES.SHIFTOUT_USED", 0)

                segmk.add_site_tag(loc, "IDDR_OR_ISERDES.IN_USE", 0)
                segmk.add_site_tag(loc, "ISERDES.IN_USE", 0)
                segmk.add_site_tag(loc, "IDDR.IN_USE", 0)

                segmk.add_site_tag(loc, "ISERDES.MODE.MASTER", 0)
                segmk.add_site_tag(loc, "ISERDES.MODE.SLAVE", 0)

                for i in iface_types:
                    if i == "NETWORKING":
                        for j in data_rates:
                            for k in data_widths[j]:
                                tag = "ISERDES.%s.%s.W%s" % (i, j, k)
                                segmk.add_site_tag(loc, tag, 0)
                    else:
                        segmk.add_site_tag(loc, "ISERDES.%s.DDR.W4" % i, 0)

                segmk.add_site_tag(loc, "ISERDES.NUM_CE.N1", 0)
                segmk.add_site_tag(loc, "ISERDES.NUM_CE.N2", 0)

                for i in range(1, 4 + 1):
                    segmk.add_site_tag(loc, "IFF.ZINIT_Q%d" % i, 0)

                for i in range(1, 4 + 1):
                    segmk.add_site_tag(loc, "IFF.ZSRVAL_Q%d" % i, 0)

    #            segmk.add_site_tag(loc, "ISERDES.IS_CLKB_INVERTED", 0)
    #            segmk.add_site_tag(loc, "ISERDES.IS_CLK_INVERTED", 1)

                segmk.add_site_tag(loc, "ISERDES.DYN_CLKDIV_INV_EN", 0)
                segmk.add_site_tag(loc, "ISERDES.DYN_CLK_INV_EN", 0)

                segmk.add_site_tag(loc, "IFFDELMUXE3.P0", 0)
                segmk.add_site_tag(loc, "IFFDELMUXE3.P1", 1)
                segmk.add_site_tag(loc, "IDELMUXE3.P0", 0)
                segmk.add_site_tag(loc, "IDELMUXE3.P1", 1)

                segmk.add_site_tag(loc, "ISERDES.OFB_USED", 0)

            # Site used as ISERDESE2
            elif verilog.unquote(params["BEL_TYPE"]) == "ISERDESE2":

                segmk.add_site_tag(loc, "IDDR_OR_ISERDES.IN_USE", 1)
                segmk.add_site_tag(loc, "ISERDES.IN_USE", 1)

                if "SHIFTOUT_USED" in params:
                    if params["CHAINED"]:
                        value = params["SHIFTOUT_USED"]
                        segmk.add_site_tag(loc, "ISERDES.SHIFTOUT_USED", value)

                if "SERDES_MODE" in params:
                    value = verilog.unquote(params["SERDES_MODE"])
                    if value == "MASTER":
                        segmk.add_site_tag(loc, "ISERDES.MODE.MASTER", 1)
                        segmk.add_site_tag(loc, "ISERDES.MODE.SLAVE", 0)
                    if value == "SLAVE":
                        segmk.add_site_tag(loc, "ISERDES.MODE.MASTER", 0)
                        segmk.add_site_tag(loc, "ISERDES.MODE.SLAVE", 1)

                iface_type = verilog.unquote(params["INTERFACE_TYPE"])
                data_rate = verilog.unquote(params["DATA_RATE"])
                data_width = int(params["DATA_WIDTH"])

                for i in iface_types:
                    if i == "NETWORKING":
                        for j in data_rates:
                            for k in data_widths[j]:
                                tag = "ISERDES.%s.%s.W%s" % (i, j, k)

                                if i == iface_type:
                                    if j == data_rate:
                                        if k == data_width:
                                            segmk.add_site_tag(loc, tag, 1)
                    else:
                        if i == iface_type:
                            segmk.add_site_tag(loc, "ISERDES.%s.DDR.W4" % i, 1)

                if "NUM_CE" in params:
                    value = params["NUM_CE"]
                    if value == 1:
                        segmk.add_site_tag(loc, "ISERDES.NUM_CE.N1", 1)
                        segmk.add_site_tag(loc, "ISERDES.NUM_CE.N2", 0)
                    if value == 2:
                        segmk.add_site_tag(loc, "ISERDES.NUM_CE.N1", 0)
                        segmk.add_site_tag(loc, "ISERDES.NUM_CE.N2", 1)

                for i in range(1, 4 + 1):
                    if ("INIT_Q%d" % i) in params:
                        segmk.add_site_tag(
                            loc, "IFF.ZINIT_Q%d" % i,
                            not params["INIT_Q%d" % i])

                for i in range(1, 4 + 1):
                    if ("SRVAL_Q%d" % i) in params:
                        segmk.add_site_tag(
                            loc, "IFF.ZSRVAL_Q%d" % i,
                            not params["SRVAL_Q%d" % i])

                for inv in ["CLK", "CLKB", "OCLK", "OCLKB", "CLKDIV",
                            "CLKDIVP"]:
                    if "IS_{}_INVERTED".format(inv) in params:
                        segmk.add_site_tag(
                            loc, "ISERDES.INV_{}".format(inv),
                            params["IS_{}_INVERTED".format(inv)])
                        segmk.add_site_tag(
                            loc, "ISERDES.ZINV_{}".format(inv),
                            not params["IS_{}_INVERTED".format(inv)])

                if "DYN_CLKDIV_INV_EN" in params:
                    value = verilog.unquote(params["DYN_CLKDIV_INV_EN"])
                    segmk.add_site_tag(
                        loc, "ISERDES.DYN_CLKDIV_INV_EN", int(value == "TRUE"))
                if "DYN_CLK_INV_EN" in params:
                    value = verilog.unquote(params["DYN_CLK_INV_EN"])
                    segmk.add_site_tag(
                        loc, "ISERDES.DYN_CLK_INV_EN", int(value == "TRUE"))

                # This parameter actually controls muxes used both in ILOGIC and
                # ISERDES mode.
                if "IOBDELAY" in params:
                    value = verilog.unquote(params["IOBDELAY"])
                    if value == "NONE":
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P0", 0)
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P1", 1)
                        segmk.add_site_tag(loc, "IDELMUXE3.P0", 0)
                        segmk.add_site_tag(loc, "IDELMUXE3.P1", 1)
                    if value == "IBUF":
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P0", 0)
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P1", 1)
                        segmk.add_site_tag(loc, "IDELMUXE3.P0", 1)
                        segmk.add_site_tag(loc, "IDELMUXE3.P1", 0)
                    if value == "IFD":
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P0", 1)
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P1", 0)
                        segmk.add_site_tag(loc, "IDELMUXE3.P0", 0)
                        segmk.add_site_tag(loc, "IDELMUXE3.P1", 1)
                    if value == "BOTH":
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P0", 1)
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P1", 0)
                        segmk.add_site_tag(loc, "IDELMUXE3.P0", 1)
                        segmk.add_site_tag(loc, "IDELMUXE3.P1", 0)

                if "OFB_USED" in params:
                    value = verilog.unquote(params["OFB_USED"])
                    segmk.add_site_tag(
                        loc, "ISERDES.OFB_USED", int(value == "TRUE"))

            # Site used as IDDR
            elif verilog.unquote(params["BEL_TYPE"]) in ["IDDR",
                                                         "IDDR_NO_CLK"]:
                segmk.add_site_tag(loc, "IDDR_OR_ISERDES.IN_USE", 1)
                segmk.add_site_tag(loc, "IDDR.IN_USE", 1)
                segmk.add_site_tag(loc, "ISERDES.IN_USE", 0)

                if "DDR_CLK_EDGE" in params:
                    value = verilog.unquote(params["DDR_CLK_EDGE"])
                    segmk.add_site_tag(
                        loc, "IFF.DDR_CLK_EDGE.OPPOSITE_EDGE",
                        int(value == "OPPOSITE_EDGE"))
                    segmk.add_site_tag(
                        loc, "IFF.DDR_CLK_EDGE.SAME_EDGE",
                        int(value == "SAME_EDGE"))
                    segmk.add_site_tag(
                        loc, "IFF.DDR_CLK_EDGE.SAME_EDGE_PIPELINED",
                        int(value == "SAME_EDGE_PIPELINED"))

                if "SRTYPE" in params:
                    value = verilog.unquote(params["SRTYPE"])
                    if value == "ASYNC":
                        segmk.add_site_tag(loc, "IFF.SRTYPE.ASYNC", 1)
                        segmk.add_site_tag(loc, "IFF.SRTYPE.SYNC", 0)
                    if value == "SYNC":
                        segmk.add_site_tag(loc, "IFF.SRTYPE.ASYNC", 0)
                        segmk.add_site_tag(loc, "IFF.SRTYPE.SYNC", 1)

                if "IDELMUX" in params:
                    if params["IDELMUX"] == 1:
                        segmk.add_site_tag(loc, "IDELMUXE3.P0", 1)
                        segmk.add_site_tag(loc, "IDELMUXE3.P1", 0)
                    else:
                        segmk.add_site_tag(loc, "IDELMUXE3.P0", 0)
                        segmk.add_site_tag(loc, "IDELMUXE3.P1", 1)

                if "IFFDELMUX" in params:
                    if params["IFFDELMUX"] == 1:
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P0", 1)
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P1", 0)
                    else:
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P0", 0)
                        segmk.add_site_tag(loc, "IFFDELMUXE3.P1", 1)

                for inv in ["C", "D"]:
                    if "IS_{}_INVERTED".format(inv) in params:
                        segmk.add_site_tag(
                            loc, "INV_{}".format(inv),
                            params["IS_{}_INVERTED".format(inv)])
                        segmk.add_site_tag(
                            loc, "ZINV_{}".format(inv),
                            not params["IS_{}_INVERTED".format(inv)])

                segmk.add_site_tag(loc, "ISERDES.NUM_CE.N1", 1)
                segmk.add_site_tag(loc, "ISERDES.NUM_CE.N2", 0)

            # Should not happen
            else:
                print("Unknown BEL_TYPE '{}'".format(params["BEL_TYPE"]))
                exit(-1)

    # Write segments and tags for later check
    def_tags = {t: 0 for d in segmk.site_tags.values() for t in d.keys()}

    with open("tags.json", "w") as fp:
        tags = {}
        for l, d in segmk.site_tags.items():
            d1 = dict(def_tags)
            d1.update({k: int(v) for k, v in d.items()})
            tags[loc_to_tile_site_map[l]] = d1

        json.dump(tags, fp, sort_keys=True, indent=1)

    def bitfilter(frame_idx, bit_idx):
        if frame_idx < 26 or frame_idx > 29:
            return False
        return True

    segmk.compile(bitfilter=bitfilter)
    segmk.write()


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import os
import random
random.seed(int(os.getenv("SEED"), 16))
from prjxray import util
from prjxray.db import Database


def gen_sites():
    db = Database(util.get_db_root(), util.get_part())
    grid = db.grid()
    for tile_name in sorted(grid.tiles()):
        loc = grid.loc_of_tilename(tile_name)
        gridinfo = grid.gridinfo_at_loc(loc)
        sites = []
        for site, site_type in gridinfo.sites.items():
            if site_type == 'BUFHCE':
                sites.append(site)

        if sites:
            yield tile_name, sorted(sites)


def write_params(params):
    pinstr = 'tile,val,site\n'
    for tile, (site, val) in sorted(params.items()):
        pinstr += '%s,%s,%s\n' % (tile, val, site)
    open('params.csv', 'w').write(pinstr)


def run():
    print('''
module top();
    ''')

    params = {}

    sites = list(gen_sites())
    for (tile_name, sites), isone in zip(sites,
                                         util.gen_fuzz_states(len(sites))):
        site_name = sites[0]
        params[tile_name] = (site_name, isone)

        print(
            '''
            (* KEEP, DONT_TOUCH, LOC = "{site}" *)
            BUFHCE #(
                .INIT_OUT({isone})
                ) buf_{site} ();
'''.format(
                site=site_name,
                isone=isone,
            ))

    print("endmodule")
    write_params(params)


if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import json
import argparse


def main():
    parser = argparse.ArgumentParser(description="")

    parser.add_argument('--error_nodes', default='output/error_nodes.json')
    parser.add_argument('--output_ignore_list', action='store_true')

    args = parser.parse_args()

    with open(args.error_nodes) as f:
        flat_error_nodes = json.load(f)

    error_nodes = {}
    for node, raw_node, generated_nodes in flat_error_nodes:
        if node not in error_nodes:
            error_nodes[node] = {
                'raw_node': set(raw_node),
                'generated_nodes': set(),
            }

        assert error_nodes[node]['raw_node'] == set(raw_node)
        error_nodes[node]['generated_nodes'].add(
            tuple(sorted(generated_nodes)))

    ignored_wires = set()

    for node, error in error_nodes.items():
        combined_generated_nodes = set()
        for generated_node in error['generated_nodes']:
            combined_generated_nodes |= set(generated_node)

        # Make sure there are not extra wires in nodes.
        assert error['raw_node'] == combined_generated_nodes, (node, error)

        good_node = max(error['generated_nodes'], key=lambda x: len(x))
        bad_nodes = error['generated_nodes'] - set((good_node, ))

        if args.output_ignore_list:
            for generated_node in bad_nodes:
                for wire in generated_node:
                    ignored_wires.add(wire)

            continue

        if max(len(generated_node) for generated_node in bad_nodes) > 1:
            assert False, node
        else:
            not_pcie = False
            for generated_node in bad_nodes:
                for wire in generated_node:
                    if not wire.startswith('PCIE'):
                        not_pcie = True
            if not_pcie:
                #print(node, good_node, map(tuple, bad_nodes))
                print(repr((node, tuple(map(tuple, bad_nodes)))))
                pass
            else:
                #print(repr((node, map(tuple, bad_nodes))))
                pass

    if args.output_ignore_list:
        for wire in ignored_wires:
            print(wire)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
from __future__ import print_function
import prjxray.db
import prjxray.lib
import argparse
import datetime
import progressbar
import multiprocessing
import pyjson5 as json5
import json
import sys
from prjxray.util import OpenSafeFile, db_root_arg, part_arg


def full_wire_name(wire_in_grid):
    return '{}/{}'.format(wire_in_grid.tile, wire_in_grid.wire)


def make_connection(wires, connection):
    wire_a = full_wire_name(connection.wire_a)
    wire_b = full_wire_name(connection.wire_b)

    if wire_a not in wires:
        wires[wire_a] = set((wire_a, ))

    if wire_b not in wires:
        wires[wire_b] = set((wire_b, ))

    wire_a_set = wires[wire_a]
    wire_b_set = wires[wire_b]

    if wire_a_set is wire_b_set:
        return

    wire_a_set |= wire_b_set

    for wire in wire_a_set:
        wires[wire] = wire_a_set


def make_connections(db_root, part):
    db = prjxray.db.Database(db_root, part)
    c = db.connections()

    wires = {}
    for connection in c.get_connections():
        make_connection(wires, connection)

    nodes = {}

    for wire_node in wires.values():
        nodes[id(wire_node)] = wire_node

    return nodes.values()


def read_json5(fname):
    with OpenSafeFile(fname, 'r') as f:
        return json5.load(f)


def main():
    parser = argparse.ArgumentParser(
        description="Tests database against raw node list.")

    db_root_arg(parser)
    part_arg(parser)
    parser.add_argument('--raw_node_root', required=True)
    parser.add_argument('--error_nodes', default="error_nodes.json")
    parser.add_argument('--ignored_wires')

    args = parser.parse_args()

    processes = min(multiprocessing.cpu_count(), 10)

    print('{} Running {} processes'.format(datetime.datetime.now(), processes))
    pool = multiprocessing.Pool(processes=processes)
    print(
        '{} Reading raw data index'.format(datetime.datetime.now(), processes))
    _, nodes = prjxray.lib.read_root_csv(args.raw_node_root)
    print('{} Reading raw_node_data'.format(datetime.datetime.now()))
    raw_node_data = []
    with progressbar.ProgressBar(max_value=len(nodes)) as bar:
        for idx, node in enumerate(pool.imap_unordered(
                read_json5,
                nodes,
                chunksize=20,
        )):
            bar.update(idx)
            raw_node_data.append(
                (node['node'], tuple(wire['wire'] for wire in node['wires'])))
            bar.update(idx + 1)

    print('{} Creating connections'.format(datetime.datetime.now()))
    generated_nodes = make_connections(args.db_root, args.part)

    print('{} Verifying connections'.format(datetime.datetime.now()))
    error_nodes = []
    prjxray.lib.verify_nodes(raw_node_data, generated_nodes, error_nodes)

    if len(error_nodes) > 0:
        if args.ignored_wires:
            with OpenSafeFile(args.ignored_wires, 'r') as f:
                ignored_wires = [l.strip() for l in f.readlines()]

        print(
            '{} Found {} errors, writing errors to {}'.format(
                datetime.datetime.now(),
                len(error_nodes),
                args.error_nodes,
            ))

        with OpenSafeFile(args.error_nodes, 'w') as f:
            json.dump(error_nodes, f, indent=2)

        if not args.ignored_wires:
            sys.exit(1)

        if not prjxray.lib.check_errors(error_nodes, ignored_wires):
            print(
                '{} Errors were not ignored via ignored_wires {}'.format(
                    datetime.datetime.now(),
                    args.ignored_wires,
                ))
            sys.exit(1)
        else:
            print(
                '{} All errors were via ignored_wires {}'.format(
                    datetime.datetime.now(),
                    args.ignored_wires,
                ))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
import fcntl
import math
import os
import random
import re
import signal
import yaml
from .roi import Roi


def timeout_handler(signum, frame):
    raise Exception("ERROR TIMEOUT: could not lock file")


class OpenSafeFile:
    """
    Opens a file in a thread-safe mode, allowing for safe read and writes
    to a file that can potentially be modified by multiple processes at
    the same time.
    """

    def __init__(self, name, mode="r", timeout=10):
        self.name = name
        self.mode = mode
        self.timeout = timeout

        self.fd = None

    def __enter__(self):
        self.fd = open(self.name, self.mode)
        self.lock_file()
        return self.fd

    def __exit__(self, exc_type, exc_value, traceback):
        self.unlock_file()
        self.fd.close()

    def lock_file(self):
        assert self.fd is not None
        try:
            signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(self.timeout)
            fcntl.flock(self.fd.fileno(), fcntl.LOCK_EX)
            signal.alarm(0)
        except Exception as e:
            print(f"{e}: {self.name}")
            exit(1)

    def unlock_file(self):
        assert self.fd is not None
        fcntl.flock(self.fd.fileno(), fcntl.LOCK_UN)


def get_db_root():
    # Used during tilegrid db bootstrap
    ret = os.getenv("XRAY_DATABASE_ROOT", None)
    if ret:
        return ret

    return "%s/%s" % (
        os.getenv("XRAY_DATABASE_DIR"), os.getenv("XRAY_DATABASE"))


def get_part():
    ret = os.getenv("XRAY_PART", None)

    return ret


def get_fabric():
    ret = os.getenv("XRAY_FABRIC", None)

    return ret


def get_part_information(db_root, part):
    filename = os.path.join(db_root, "mapping", "parts.yaml")
    assert os.path.isfile(filename), \
        "Mapping file {} does not exist".format(filename)
    with open(filename, 'r') as stream:
        part_mapping = yaml.load(stream, Loader=yaml.FullLoader)
    part = part_mapping.get(part, None)
    assert part, "Part {} not found in {}".format(part, part_mapping)
    return part


def set_part_information(db_root, information):
    filename = os.path.join(db_root, "mapping", "parts.yaml")
    with OpenSafeFile(filename, 'w+') as stream:
        yaml.dump(information, stream)
    assert os.path.isfile(filename), \
        "Mapping file {} does not exist".format(filename)


def get_part_resources(file_path, part):
    filename = os.path.join(file_path, "resources.yaml")
    assert os.path.isfile(filename), \
        "Mapping file {} does not exist".format(filename)
    with open(filename, 'r') as stream:
        res_mapping = yaml.load(stream, Loader=yaml.FullLoader)
    res = res_mapping.get(part, None)
    assert res, "Part {} not found in {}".format(part, part_mapping)
    return res


def set_part_resources(file_path, information):
    filename = os.path.join(file_path, "resources.yaml")
    with OpenSafeFile(filename, 'w+') as stream:
        yaml.dump(information, stream)
    assert os.path.isfile(filename), \
        "Mapping file {} does not exist".format(filename)


def get_fabric_for_part(db_root, part):
    filename = os.path.join(db_root, "mapping", "devices.yaml")
    assert os.path.isfile(filename), \
        "Mapping file {} does not exist".format(filename)
    part = get_part_information(db_root, part)
    with OpenSafeFile(filename, 'r') as stream:
        device_mapping = yaml.load(stream, Loader=yaml.FullLoader)
    device = device_mapping.get(part['device'], None)
    assert device, "Device {} not found in {}".format(
        part['device'], device_mapping)
    return device['fabric']


def get_devices(db_root):
    filename = os.path.join(db_root, "mapping", "devices.yaml")
    assert os.path.isfile(filename), \
        "Mapping file {} does not exist".format(filename)
    with open(filename, 'r') as stream:
        device_mapping = yaml.load(stream, Loader=yaml.FullLoader)
    return device_mapping


def get_parts(db_root):
    filename = os.path.join(db_root, "mapping", "parts.yaml")
    assert os.path.isfile(filename), \
        "Mapping file {} does not exist".format(filename)
    with open(filename, 'r') as stream:
        part_mapping = yaml.load(stream, Loader=yaml.FullLoader)
    return part_mapping


def roi_xy():
    x1 = int(os.getenv('XRAY_ROI_GRID_X1', 0))
    x2 = int(os.getenv('XRAY_ROI_GRID_X2', 58))
    y1 = int(os.getenv('XRAY_ROI_GRID_Y1', 0))
    y2 = int(os.getenv('XRAY_ROI_GRID_Y2', 52))

    return (x1, x2), (y1, y2)


def create_xy_fun(prefix):
    """ Create function that extracts X and Y coordinate from a prefixed string

    >>> fun = create_xy_fun(prefix='')
    >>> fun('X5Y23')
    (5, 23)
    >>> fun('X0Y0')
    (0, 0)
    >>> fun('X50Y100')
    (50, 100)

    >>> fun = create_xy_fun(prefix='SITE_')
    >>> fun('SITE_X5Y23')
    (5, 23)
    >>> fun('SITE_X0Y0')
    (0, 0)
    >>> fun('SITE_X50Y100')
    (50, 100)

    """
    compiled_re = re.compile(
        '^{prefix}X([0-9]+)Y([0-9]+)$'.format(prefix=prefix))

    def get_xy(s):
        m = compiled_re.match(s)
        assert m, (prefix, s)

        x = int(m.group(1))
        y = int(m.group(2))

        return x, y

    return get_xy


def slice_xy():
    '''Return (X1, X2), (Y1, Y2) from XRAY_ROI, exclusive end (for xrange)'''
    # SLICE_X12Y100:SLICE_X27Y149
    # Note XRAY_ROI_GRID_* is something else
    m = re.match(
        r'SLICE_X([0-9]*)Y([0-9]*):SLICE_X([0-9]*)Y([0-9]*)',
        os.getenv('XRAY_ROI'))
    ms = [int(m.group(i + 1)) for i in range(4)]
    return ((ms[0], ms[2] + 1), (ms[1], ms[3] + 1))


def get_roi():
    from .db import Database
    (x1, x2), (y1, y2) = roi_xy()
    db = Database(get_db_root(), get_part())
    return Roi(db=db, x1=x1, x2=x2, y1=y1, y2=y2)


def gen_sites_xy(site_types):
    for _tile_name, site_name, _site_type in get_roi().gen_sites(site_types):
        m = re.match(r'.*_X([0-9]*)Y([0-9]*)', site_name)
        x, y = int(m.group(1)), int(m.group(2))
        yield (site_name, (x, y))


def site_xy_minmax(site_types):
    '''Return (X1, X2), (Y1, Y2) from XY_ROI, exclusive end (for xrange)'''
    xmin = 9999
    xmax = -1
    ymin = 9999
    ymax = -1
    for _site_name, (x, y) in gen_sites_xy(site_types):
        xmin = min(xmin, x)
        xmax = max(xmax, x)
        ymin = min(ymin, y)
        ymax = max(ymax, y)
    return (xmin, xmax + 1), (ymin, ymax + 1)


# we know that all bits for CLB MUXes are in frames 30 and 31, so filter all other bits
def bitfilter_clb_mux(frame_idx, bit_idx):
    return frame_idx in [30, 31]


def db_root_arg(parser):
    database_dir = os.getenv("XRAY_DATABASE_DIR")
    database = os.getenv("XRAY_DATABASE")
    db_root_kwargs = {}
    if database_dir is None or database is None:
        db_root_kwargs['required'] = True
    else:
        db_root_kwargs['required'] = False
        db_root_kwargs['default'] = os.path.join(database_dir, database)
    parser.add_argument('--db-root', help="Database root.", **db_root_kwargs)


def part_arg(parser):
    part = os.getenv("XRAY_PART")
    part_kwargs = {}
    if part is None:
        part_kwargs['required'] = True
    else:
        part_kwargs['required'] = False
        part_kwargs['default'] = part
    parser.add_argument(
        '--part',
        help="Part name. When not given defaults to XRAY_PART env. var.",
        **part_kwargs)


def parse_db_line(line):
    '''Return tag name, bit values (if any), mode (if any)'''
    parts = line.split()
    # Ex: CLBLL_L.SLICEL_X0.AMUX.A5Q
    assert len(parts), "Empty line"
    tag = parts[0]
    if tag == 'bit':
        raise ValueError("Wanted bits db but got mask db")
    assert re.match(r'[A-Z0-9_.]+',
                    tag), "Invalid tag name: %s, line: %s" % (tag, line)
    orig_bits = line.replace(tag + " ", "")
    if re.match(r'^origin:', orig_bits):
        origin = parts[1][7:]
        bits = frozenset(parts[2:])
        orig_bits = line.replace(tag + " " + origin + " ", "")
    else:
        origin = None
        bits = frozenset(parts[1:])
    # <0 candidates> etc
    # Ex: INT_L.BYP_BOUNCE5.BYP_ALT5 always
    if "<" in orig_bits or "always" == orig_bits:
        return tag, None, orig_bits, origin

    # Ex: CLBLL_L.SLICEL_X0.AOUTMUX.A5Q !30_06 !30_08 !30_11 30_07
    for bit in bits:
        # 19_39
        # 100_319
        assert re.match(r'[!]*[0-9]+_[0-9]+', bit), "Invalid bit: %s" % bit
    return tag, bits, None, origin


def parse_db_lines(fn):
    with OpenSafeFile(fn, "r") as f:
        for line in f:
            yield line, parse_db_line(line)


def write_db_lines(fn, entries, track_origin=False):
    new_lines = []
    for tag, (bits, origin) in entries.items():
        if track_origin:
            assert origin is not None
            new_line = " ".join([tag, "origin:" + origin] + sorted(bits))
        else:
            new_line = " ".join([tag] + sorted(bits))
        new_lines.append(new_line)

    with OpenSafeFile(fn, "w") as f:
        for line in sorted(new_lines):
            print(line, file=f)


def parse_tagbit(x):
    # !30_07
    if x[0] == '!':
        isset = False
        numstr = x[1:]
    else:
        isset = True
        numstr = x
    frame, word = numstr.split("_")
    # second part forms a tuple refereced in sets
    return (isset, (int(frame, 10), int(word, 10)))


def addr_bit2word(bitaddr):
    word = bitaddr // 32
    bit = bitaddr % 32
    return word, bit


def addr2str(addr, word, bit):
    # Make like .bits file: bit_00020b14_073_05
    # also similar to .db file: CLBLL_L.SLICEL_X0.CEUSEDMUX 01_39
    assert 0 <= bit <= 31
    return "%08x_%03u_%02u" % (addr, word, bit)


# matches lib/include/prjxray/xilinx/xc7series/block_type.h
block_type_i2s = {
    0: 'CLB_IO_CLK',
    1: 'BLOCK_RAM',
    2: 'CFG_CLB',
    # special...maybe should error until we know what it is?
    # 3: 'RESERVED',
}
block_type_s2i = {}
for k, v in block_type_i2s.items():
    block_type_s2i[v] = k


def addr2btype(base_addr):
    '''
    Convert integer address to block type

    Table 5-24: Frame Address Register Description
    Bit Index: [25:23]
    https://www.xilinx.com/support/documentation/user_guides/ug470_7Series_Config.pdf
    "Valid block types are CLB, I/O, CLK ( 000 ), block RAM content ( 001 ), and CFG_CLB ( 010 ). A normal bitstream does not include type 011 ."
    '''
    block_type_i = (base_addr >> 23) & 0x7
    return block_type_i2s[block_type_i]


def specn():
    # ex: build/specimen_001
    specdir = os.getenv("SPECDIR")
    return int(re.match(".*specimen_([0-9]*)", specdir).group(1), 10)


def gen_fuzz_choices(nvals):
    next_p2_states = 2**math.ceil(math.log(nvals, 2))
    n = next_p2_states

    full_mask = 2**next_p2_states - 1

    choices = []
    invert_choices = []

    num_or = 1
    while n > 0:
        mask = 2**n - 1

        val = 0

        for offset in range(0, num_or, 2):
            shift = offset * next_p2_states // num_or
            val |= mask << shift

        choices.append(full_mask ^ val)
        invert_choices.append(val)

        n //= 2
        num_or *= 2

    choices.extend(invert_choices)

    return choices


def gen_fuzz_states(nvals):
    '''
    Generates an optimal encoding to solve single bits as quickly as possible

    tilegrid's initial solve for 4 bits works like this:
    Initial reference value of all 0s:
    0000
    Then one-hot for each:
    0001
    0010
    0100
    1000
    Which requires 5 samples total to diff these

    However, using correlation instead its possible to resolve n bits using ceil(log(n, 2)) + 1 samples
    With 4 positions it takes only 3 samples:
    0000
    0101
    1010
    '''

    choices = gen_fuzz_choices(nvals)
    spec_idx = specn() - 1
    if spec_idx < len(choices):
        bits = choices[spec_idx]
    else:
        next_p2_states = 2**math.ceil(math.log(nvals, 2))
        bits = random.randint(0, 2**next_p2_states)

    for i in range(nvals):
        mask = (1 << i)
        yield int(bool(bits & mask))


def add_bool_arg(parser, yes_arg, default=False, **kwargs):
    dashed = yes_arg.replace('--', '')
    dest = dashed.replace('-', '_')
    parser.add_argument(
        yes_arg, dest=dest, action='store_true', default=default, **kwargs)
    parser.add_argument(
        '--no-' + dashed, dest=dest, action='store_false', **kwargs)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" Tool to cleanup site pins JSON dumps.

This tool has two behaviors.  This first is to rename site names from global
coordinates to site local coordinates.  The second is remove the tile prefix
from node names.

For example CLBLM_L_X8Y149 contains two sites named SLICE_X10Y149 and
SLICE_X11Y149. SLICE_X10Y149 becomes X0Y0 and SLICE_X11Y149 becomes X1Y0.
"""

from __future__ import print_function
import json
import json5
import re
import sys
import copy

# All site names appear to follow the pattern <type>_X<abs coord>Y<abs coord>.
# Generally speaking, only the tile relatively coordinates are required to
# assemble arch defs, so we re-origin the coordinates to be relative to the tile
# (e.g. start at X0Y0) and discard the prefix from the name.
SITE_COORDINATE_PATTERN = re.compile('^(.+)_X([0-9]+)Y([0-9]+)$')


def find_origin_coordinate(sites):
    """ Find the coordinates of each site within the tile, and then subtract the
      smallest coordinate to re-origin them all to be relative to the tile.
  """

    if len(sites) == 0:
        return 0, 0

    def inner_():
        for site in sites:
            coordinate = SITE_COORDINATE_PATTERN.match(site['name'])
            assert coordinate is not None, site

            x_coord = int(coordinate.group(2))
            y_coord = int(coordinate.group(3))
            yield x_coord, y_coord

    x_coords, y_coords = zip(*inner_())
    min_x_coord = min(x_coords)
    min_y_coord = min(y_coords)

    return min_x_coord, min_y_coord


def create_site_pin_to_wire_maps(tile_name, nodes):
    """ Create a map from site_pin names to nodes.

  Create a mapping from site pins to tile local wires.  For each node that is
  attached to a site pin, there should only be 1 tile local wire.

  """

    # Remove tile prefix (e.g. CLBLM_L_X8Y149/) from node names.
    # Routing resources will not have the prefix.
    tile_prefix = tile_name + '/'
    site_pin_to_wires = {}

    for node in nodes:
        if len(node['site_pins']) == 0:
            continue

        wire_names = [
            wire for wire in node['wires'] if wire.startswith(tile_prefix)
        ]
        assert len(wire_names) == 1, (node, tile_prefix)

        for site_pin in node["site_pins"]:
            assert site_pin not in site_pin_to_wires
            site_pin_to_wires[site_pin] = wire_names[0]

    return site_pin_to_wires


def main():
    site_pins = json5.load(sys.stdin)

    output_site_pins = {}
    output_site_pins["tile_type"] = site_pins["tile_type"]
    output_site_pins["sites"] = copy.deepcopy(site_pins["sites"])

    site_pin_to_wires = create_site_pin_to_wire_maps(
        site_pins['tile_name'], site_pins['nodes'])
    min_x_coord, min_y_coord = find_origin_coordinate(site_pins['sites'])

    for site in output_site_pins['sites']:
        orig_site_name = site['name']
        coordinate = SITE_COORDINATE_PATTERN.match(orig_site_name)

        x_coord = int(coordinate.group(2))
        y_coord = int(coordinate.group(3))
        site['name'] = 'X{}Y{}'.format(
            x_coord - min_x_coord, y_coord - min_y_coord)
        site['prefix'] = coordinate.group(1)
        site['x_coord'] = x_coord - min_x_coord
        site['y_coord'] = y_coord - min_y_coord

        for site_pin in site['site_pins']:
            assert site_pin['name'].startswith(orig_site_name + '/')
            if site_pin['name'] in site_pin_to_wires:
                site_pin['wire'] = site_pin_to_wires[site_pin['name']]
            else:
                print(
                    (
                        '***WARNING***: Site pin {} for tile type {} is not connected, '
                        'make sure all instaces of this tile type has this site_pin '
                        'disconnected.').format(
                            site_pin['name'], site_pins['tile_type']),
                    file=sys.stderr)

            site_pin['name'] = site_pin['name'][len(orig_site_name) + 1:]

    json.dumps(output_site_pins, indent=2, sort_keys=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
""" This takes a JSON file generated with write_timing_info.tcl and generates
a spreadsheet with the prjxray timing model and compares it with the
interconnect timing output from Vivado.

"""
import argparse
import json
from openpyxl import Workbook, utils
from prjxray.tile import OutPinTiming, InPinTiming
from prjxray.timing import Outpin, Inpin, Wire, Buffer, \
        PassTransistor, IntristicDelay, RcElement, PvtCorner
from prjxray.math_models import ExcelMathModel
from prjxray.db import Database
from prjxray.util import OpenSafeFile
from prjxray import util

FAST = PvtCorner.FAST
SLOW = PvtCorner.SLOW


class TimingLookup(object):
    def __init__(self, db, nodes):
        self.db = db
        self.grid = db.grid()
        self.nodes = nodes

    def try_find_site_pin(self, site_pin_node, node_idx):
        site_pin_wire = self.nodes[site_pin_node]['wires'][node_idx]['name']
        tile, wire_in_tile = site_pin_wire.split('/')

        gridinfo = self.grid.gridinfo_at_tilename(tile)

        tile_type = self.db.get_tile_type(gridinfo.tile_type)

        for site in tile_type.get_sites():
            for site_pin in site.site_pins:
                if site_pin.wire == wire_in_tile:
                    return site_pin

        return None

    def find_site_pin(self, site_pin_node, node_idx):
        site_pin = self.try_find_site_pin(site_pin_node, node_idx)

        assert site_pin is not None, site_pin_node
        return site_pin

    def find_pip(self, pip_name):
        tile, pip = pip_name.split('/')

        gridinfo = self.grid.gridinfo_at_tilename(tile)

        tile_type = self.db.get_tile_type(gridinfo.tile_type)

        return tile_type.get_pip_by_name(pip)

    def find_wire(self, wire_name):
        tile, wire_in_tile = wire_name.split('/')

        gridinfo = self.grid.gridinfo_at_tilename(tile)

        tile_type = self.db.get_tile_type(gridinfo.tile_type)

        return tile_type.wires[wire_in_tile]


def delays_to_cells(ws, row, delays, cells):
    cells['FAST_MAX'] = 'E{}'.format(row)
    cells['FAST_MIN'] = 'F{}'.format(row)
    cells['SLOW_MAX'] = 'G{}'.format(row)
    cells['SLOW_MIN'] = 'H{}'.format(row)

    if delays is not None:
        ws[cells['FAST_MAX']] = delays[FAST].max
        ws[cells['FAST_MIN']] = delays[FAST].min
        ws[cells['SLOW_MAX']] = delays[SLOW].max
        ws[cells['SLOW_MIN']] = delays[SLOW].min
    else:
        ws[cells['FAST_MAX']] = 0
        ws[cells['FAST_MIN']] = 0
        ws[cells['SLOW_MAX']] = 0
        ws[cells['SLOW_MIN']] = 0


def cells_to_delays(cells):
    return {
        FAST: IntristicDelay(min=cells['FAST_MIN'], max=cells['FAST_MAX']),
        SLOW: IntristicDelay(min=cells['SLOW_MIN'], max=cells['SLOW_MAX']),
    }


class Net(object):
    def __init__(self, net):
        self.net = net
        self.ipin_nodes = {}
        self.row = None
        self.math = ExcelMathModel()
        self.models = {}

        for ipin in net['ipins']:
            for ipin_node in ipin['node'].strip().split(' '):
                self.ipin_nodes[ipin_node] = ipin

        # Map of wire name to parent node
        self.wire_to_node = {}

        # Map of node name to node
        self.node_name_to_node = {}

        for node in net['nodes']:
            self.node_name_to_node[node['name']] = node
            for wire in node['wires']:
                self.wire_to_node[wire['name']] = node

        # Map of (src node, dst wire).
        self.pips = {}
        for pip in net['pips']:
            src_node = self.wire_to_node[pip['src_wire']]['name']
            dst_wire = pip['dst_wire'].split('/')[1]
            self.pips[(src_node, dst_wire)] = pip

            if not int(pip['is_directional']):
                dst_node = self.wire_to_node[pip['dst_wire']]['name']
                src_wire = pip['src_wire'].split('/')[1]
                self.pips[(dst_node, src_wire)] = pip

    def extend_rc_tree(self, ws, current_rc_root, timing_lookup, node):
        rc_elements = []

        # LV nodes have a workaround applied because of a work around in the
        # pip timing data.
        is_lv_node = any(
            wire['name'].split('/')[1].startswith('LV')
            for wire in node['wires'])
        for idx, wire in enumerate(node['wires']):
            wire_timing = timing_lookup.find_wire(wire['name'])
            ws['A{}'.format(self.row)] = wire['name']
            ws['B{}'.format(self.row)] = 'Part of wire'

            if wire_timing is not None:
                cells = {}
                cells['R'] = 'C{}'.format(self.row)
                cells['C'] = 'D{}'.format(self.row)
                if not is_lv_node:
                    ws[cells['R']] = wire_timing.resistance
                    ws[cells['C']] = wire_timing.capacitance
                else:
                    # Only use first 2 wire RC's, ignore the rest.  It appears
                    # that some of the RC constant was lumped into the switch
                    # timing, so don't double count.
                    if idx < 2:
                        ws[cells['R']] = wire_timing.resistance
                        ws[cells['C']] = wire_timing.capacitance
                    else:
                        ws[cells['R']] = 0
                        ws[cells['C']] = 0

                rc_elements.append(
                    RcElement(
                        resistance=cells['R'],
                        capacitance=cells['C'],
                    ))

            self.row += 1

        wire_rc_node = Wire(rc_elements=rc_elements, math=self.math)
        self.models[self.row - 1] = wire_rc_node

        current_rc_root.set_sink_wire(wire_rc_node)

        return wire_rc_node

    def descend_route(
            self,
            ws,
            timing_lookup,
            current_node,
            route,
            route_idx,
            current_rc_root,
            was_opin=False):
        """ Traverse the next pip, or recurse deeper. """

        # descend_route should've consumed this token
        assert route[route_idx] != '}'

        while route[route_idx] == '{':
            # Go deeper
            route_idx = self.descend_route(
                ws,
                timing_lookup,
                current_node,
                route,
                route_idx=route_idx + 1,
                current_rc_root=current_rc_root,
                was_opin=was_opin)

        next_edge = (current_node, route[route_idx])
        route_idx += 1
        assert next_edge in self.pips, (next_edge, self.pips.keys())

        pip = self.pips[next_edge]
        is_backward = self.wire_to_node[
            pip['dst_wire']]['name'] == current_node
        if not is_backward:
            assert self.wire_to_node[
                pip['src_wire']]['name'] == current_node, (current_node, pip)

        pip_info = timing_lookup.find_pip(pip['name'])
        if not is_backward:
            pip_timing = pip_info.timing
            current_node = self.wire_to_node[pip['dst_wire']]['name']
        else:
            pip_timing = pip_info.backward_timing
            current_node = self.wire_to_node[pip['src_wire']]['name']

        ws['A{}'.format(self.row)] = pip['name']

        cells = {}
        cells['R'] = 'C{}'.format(self.row)
        cells['C'] = 'D{}'.format(self.row)
        delays_to_cells(
            ws, row=self.row, delays=pip_timing.delays, cells=cells)
        delays = cells_to_delays(cells)

        if pip_info.is_pass_transistor:
            ws['B{}'.format(self.row)] = 'PassTransistor'
            ws[cells['R']] = pip_timing.drive_resistance

            pip_model = PassTransistor(
                drive_resistance=cells['R'],
                delays=delays,
            )
        else:
            ws['B{}'.format(self.row)] = 'Buffer'
            if pip_timing.drive_resistance is not None:
                ws[cells['R']] = pip_timing.drive_resistance
                if pip_timing.drive_resistance == 0 and was_opin:
                    new_site_pin = timing_lookup.try_find_site_pin(
                        current_node, node_idx=0)
                    if new_site_pin is not None:
                        ws[cells['R']] = new_site_pin.timing.drive_resistance
            else:
                ws[cells['R']] = 0

            if pip_timing.internal_capacitance is not None:
                ws[cells['C']] = pip_timing.internal_capacitance
            else:
                ws[cells['C']] = 0

            pip_model = Buffer(
                drive_resistance=cells['R'],
                internal_capacitance=cells['C'],
                delays=cells_to_delays(cells),
            )

        self.models[self.row] = pip_model

        current_rc_root.add_child(pip_model)
        self.row += 1

        if current_node in self.ipin_nodes:
            assert route[route_idx] in ['}', 'IOB_O_OUT0', 'IOB_T_OUT0'], (
                route_idx,
                route[route_idx],
            )
            route_idx += 1

        node = self.node_name_to_node[current_node]

        current_rc_root = self.extend_rc_tree(
            ws, pip_model, timing_lookup, node)

        if current_node in self.ipin_nodes:
            ipin = self.ipin_nodes[current_node]

            cells = {}
            name = '{} to {}'.format(self.net['opin']['name'], ipin['name'])
            ws['A{}'.format(self.row)] = ipin['name']
            ws['B{}'.format(self.row)] = 'Inpin'

            site_pin = timing_lookup.find_site_pin(current_node, node_idx=-1)
            assert isinstance(site_pin.timing, InPinTiming)

            cells = {}
            cells['C'] = 'D{}'.format(self.row)
            delays_to_cells(
                ws, row=self.row, delays=site_pin.timing.delays, cells=cells)
            delays = cells_to_delays(cells)

            ws[cells['C']] = site_pin.timing.capacitance
            ipin_model = Inpin(
                capacitance=cells['C'], delays=delays, name=name)
            self.models[self.row] = ipin_model
            current_rc_root.add_child(ipin_model)
            self.row += 1

            #Sum delays only (sum*1000)
            #Sum delays + capacitive delay
            #
            #Total delay (from Vivado)
            ws['A{}'.format(self.row)] = '{}: {} sum delays'.format(
                self.net['net'], name)
            self.row += 1

            ws['A{}'.format(self.row)] = '{}: {}sum delays + cap delay'.format(
                self.net['net'], name)
            self.row += 2

            ws['A{}'.format(
                self.row)] = '{}: {}Total delay (from Vivado)'.format(
                    self.net['net'], name)

            ws['E{}'.format(self.row)] = ipin['ic_delays']['FAST_MAX']
            ws['F{}'.format(self.row)] = ipin['ic_delays']['FAST_MIN']
            ws['G{}'.format(self.row)] = ipin['ic_delays']['SLOW_MAX']
            ws['H{}'.format(self.row)] = ipin['ic_delays']['SLOW_MIN']

            self.row += 2

            return route_idx
        else:
            return self.descend_route(
                ws,
                timing_lookup,
                current_node,
                route,
                route_idx=route_idx,
                current_rc_root=current_rc_root)

    def walk_route(self, ws, timing_lookup):
        """ Walk route, creating rows in table.

        First row will always be the OPIN, followed by the node/wire connected
        to the OPIN.  After a node/wire is always 1 or more pips. After a pip
        is always a node/wire.  A terminal node/wire will then reach an IPIN.
        """

        self.row = 2
        ws['A{}'.format(self.row)] = self.net['opin']['wire']

        site_pin = timing_lookup.find_site_pin(
            self.net['opin']['node'], node_idx=0)
        assert isinstance(site_pin.timing, OutPinTiming)

        ws['B{}'.format(self.row)] = 'Outpin'
        ws['C{}'.format(self.row)] = site_pin.timing.drive_resistance

        cells = {}
        cells['R'] = 'C{}'.format(self.row)

        delays_to_cells(
            ws, row=self.row, delays=site_pin.timing.delays, cells=cells)

        model_root = Outpin(
            resistance=cells['R'], delays=cells_to_delays(cells))
        self.models[self.row] = model_root
        self.row += 1

        node = self.net['opin']['node']

        tile, first_wire = self.net['opin']['node'].split('/')

        route = [r for r in self.net['route'].strip().split(' ') if r != '']
        assert route[0] == '{'
        assert route[1] == first_wire

        node = self.node_name_to_node[node]
        current_rc_root = self.extend_rc_tree(
            ws, model_root, timing_lookup, node)

        self.descend_route(
            ws,
            timing_lookup,
            node['name'],
            route,
            route_idx=2,
            current_rc_root=current_rc_root,
            was_opin=True)

        model_root.propigate_delays(self.math)

        model_rows = {}

        for row, model in self.models.items():
            model_rows[id(model)] = row

        for row, model in self.models.items():
            rc_delay = model.get_rc_delay()
            if rc_delay is not None:
                ws['J{}'.format(row)] = self.math.eval(rc_delay)

            downstream_cap = model.get_downstream_cap()
            if downstream_cap is not None:
                ws['I{}'.format(row)] = self.math.eval(downstream_cap)

            if isinstance(model, Inpin):
                ipin_results = {
                    'Name': model.name,
                    'truth': {},
                    'computed': {},
                }

                ipin_delays = {}

                DELAY_COLS = (
                    ('E', 'FAST_MAX'),
                    ('F', 'FAST_MIN'),
                    ('G', 'SLOW_MAX'),
                    ('H', 'SLOW_MIN'),
                )

                for col, value in DELAY_COLS:
                    ipin_delays[value] = []
                    ipin_results['computed'][
                        value] = '{title}!{col}{row}'.format(
                            title=utils.quote_sheetname(ws.title),
                            col=col,
                            row=row + 2)
                    ipin_results['truth'][value] = '{title}!{col}{row}'.format(
                        title=utils.quote_sheetname(ws.title),
                        col=col,
                        row=row + 4)

                rc_delays = []

                for model in model.get_delays():
                    delays = model.get_intrinsic_delays()
                    if delays is not None:
                        ipin_delays['FAST_MAX'].append(delays[FAST].max)
                        ipin_delays['FAST_MIN'].append(delays[FAST].min)
                        ipin_delays['SLOW_MAX'].append(delays[SLOW].max)
                        ipin_delays['SLOW_MIN'].append(delays[SLOW].min)

                    if id(model) in model_rows:
                        rc_delays.append('J{}'.format(model_rows[id(model)]))

                ws['J{}'.format(row + 1)] = self.math.eval(
                    self.math.sum(rc_delays))

                for col, value in DELAY_COLS:
                    ws['{}{}'.format(col, row + 1)] = self.math.eval(
                        self.math.sum(ipin_delays[value]))
                    ws['{}{}'.format(
                        col, row + 2)] = '=1000*({col}{row} + J{row})'.format(
                            col=col, row=row + 1)

                yield ipin_results


def add_net(wb, net, timing_lookup):
    replace_underscore = str.maketrans('[]\\:/', '_____')
    ws = wb.create_sheet(
        title="Net {}".format(net['net'].translate(replace_underscore)))

    # Header
    ws['A1'] = 'Name'
    ws['B1'] = 'Type'
    ws['C1'] = 'RES'
    ws['D1'] = 'CAP'
    ws['E1'] = 'FAST_MAX'
    ws['F1'] = 'FAST_MIN'
    ws['G1'] = 'SLOW_MAX'
    ws['H1'] = 'SLOW_MIN'
    ws['I1'] = 'Downstream C'
    ws['J1'] = 'Delay from cap'

    net_obj = Net(net)
    yield from net_obj.walk_route(ws, timing_lookup)


def build_wire_filter(wire_filter):
    wires_to_include = set()

    with OpenSafeFile(wire_filter) as f:
        for l in f:
            wire = l.strip()
            if not wire:
                continue
            wires_to_include.add(wire)

    def filter_net(net):
        wires_in_net = set()

        for node in net['nodes']:
            for wire in node['wires']:
                wires_in_net.add(wire['name'])

        return len(wires_in_net & wires_to_include) > 0

    return filter_net


def main():
    parser = argparse.ArgumentParser(
        description="Create timing worksheet for 7-series timing analysis.")

    util.db_root_arg(parser)
    util.part_arg(parser)
    parser.add_argument('--timing_json', required=True)
    parser.add_argument('--output_xlsx', required=True)
    parser.add_argument(
        '--wire_filter',
        help='List of wires that must be present in a net to be output')

    args = parser.parse_args()

    with OpenSafeFile(args.timing_json) as f:
        timing = json.load(f)

    db = Database(args.db_root, args.part)

    nodes = {}
    for net in timing:
        for node in net['nodes']:
            nodes[node['name']] = node

    timing_lookup = TimingLookup(db, nodes)

    wb = Workbook()
    summary_ws = wb[wb.sheetnames[0]]
    summary_ws.title = 'Summary'

    summary_ws['A1'] = 'Name'

    cols = ['FAST_MAX', 'FAST_MIN', 'SLOW_MAX', 'SLOW_MIN']
    cur_col = 'B'
    for col in cols:
        summary_ws['{}1'.format(cur_col)] = col
        cur_col = chr(ord(cur_col) + 1)
        summary_ws['{}1'.format(cur_col)] = 'Computed ' + col
        cur_col = chr(ord(cur_col) + 3)

    if args.wire_filter:
        wire_filter = build_wire_filter(args.wire_filter)
    else:
        wire_filter = lambda x: True

    summary_row = 2

    timing = [net for net in timing if wire_filter(net)]
    for idx, net in enumerate(timing):
        if '<' in net['route']:
            print(
                "WARNING: Skipping net {} because it has complicated route description."
                .format(net['net']))
            continue

        print('Process net {} ({} / {})'.format(net['net'], idx, len(timing)))
        for summary_cells in add_net(wb, net, timing_lookup):
            summary_ws['A{}'.format(summary_row)] = summary_cells['Name']

            cur_col = 'B'
            for col in cols:
                truth_col = chr(ord(cur_col) + 0)
                computed_col = chr(ord(cur_col) + 1)
                error_col = chr(ord(cur_col) + 2)
                error_per_col = chr(ord(cur_col) + 3)
                summary_ws['{}{}'.format(
                    truth_col,
                    summary_row)] = '=' + summary_cells['truth'][col]
                summary_ws['{}{}'.format(
                    computed_col,
                    summary_row)] = '=' + summary_cells['computed'][col]
                summary_ws['{}{}'.format(
                    error_col,
                    summary_row)] = '={truth}{row}-{comp}{row}'.format(
                        truth=truth_col, comp=computed_col, row=summary_row)
                summary_ws['{}{}'.format(
                    error_per_col,
                    summary_row)] = '={error}{row}/{truth}{row}'.format(
                        error=error_col, truth=truth_col, row=summary_row)

                cur_col = chr(ord(cur_col) + 4)

            summary_row += 1

    wb.save(filename=args.output_xlsx)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

import os, re


def maketodo(pipfile, dbfile):
    todos = set()
    with open(pipfile, "r") as f:
        for line in f:
            line = line.split()
            todos.add(line[0])
    with open(dbfile, "r") as f:
        for line in f:
            line = line.split()
            todos.remove(line[0])
    for line in todos:
        if line.endswith(".GND_WIRE") or line.endswith(".VCC_WIRE"):
            print(line)


maketodo(
    "pips_int_l.txt", "%s/%s/segbits_int_l.db" %
    (os.getenv("XRAY_DATABASE_DIR"), os.getenv("XRAY_DATABASE")))
maketodo(
    "pips_int_r.txt", "%s/%s/segbits_int_r.db" %
    (os.getenv("XRAY_DATABASE_DIR"), os.getenv("XRAY_DATABASE")))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC
'''
Pretty print FASM.

Sanity checks FASM against prjxray database.
Can output canonical FASM.
In the future may support other formatting options.

'''

import os
import fasm
from prjxray import db


def process_fasm(db_root, part, fasm_file, canonical):
    database = db.Database(db_root, part)
    grid = database.grid()

    for fasm_line in fasm.parse_fasm_filename(fasm_file):
        if not fasm_line.set_feature:
            if not canonical:
                yield fasm_line

        for feature in fasm.canonical_features(fasm_line.set_feature):
            parts = feature.feature.split('.')
            tile = parts[0]

            gridinfo = grid.gridinfo_at_tilename(tile)
            tile_segbits = database.get_tile_segbits(gridinfo.tile_type)

            address = 0
            if feature.start is not None:
                address = feature.start

            feature_name = '{}.{}'.format(
                gridinfo.tile_type, '.'.join(parts[1:]))

            # Convert feature to bits.  If no bits are set, feature is
            # psuedo pip, and should not be output from canonical FASM.
            bits = tuple(
                tile_segbits.feature_to_bits(feature_name, address=address))
            if len(bits) == 0 and canonical:
                continue

            # In canonical output, only output the canonical features.
            if canonical:
                yield fasm.FasmLine(
                    set_feature=feature,
                    annotations=None,
                    comment=None,
                )

        # If not in canonical mode, output original FASM line
        if not canonical:
            yield fasm_line


def run(db_root, part, fasm_file, canonical):
    print(
        fasm.fasm_tuple_to_string(
            process_fasm(db_root, part, fasm_file, canonical),
            canonical=canonical))


def main():
    import argparse

    parser = argparse.ArgumentParser(description='Pretty print a FASM file.')

    util.db_root_arg(parser)
    util.part_arg(parser)
    parser.add_argument('fasm_file', help='Input FASM file')
    parser.add_argument(
        '--canonical', help='Output canonical bitstream.', action='store_true')
    args = parser.parse_args()

    run(args.db_root, args.part, args.fasm_file, args.canonical)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2017-2020  The Project X-Ray Authors.
#
# Use of this source code is governed by a ISC-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/ISC
#
# SPDX-License-Identifier: ISC

from prjxray import bitstream


def write(bits_fn, fnout, tags):
    '''
    seg 00020000_046
    bit 18_20
    bit 39_63
    tag LIOB33.IOB_Y1.REFBIT 0
    '''
    fout = open(fnout, "w")

    def line(s):
        fout.write(s + "\n")

    # Everything relative to start of bitstream
    line("seg 00000000_000")

    bitdata = bitstream.load_bitdata2(open(bits_fn, "r"))

    for frame, words in bitdata.items():
        for word, wbits in words.items():
            for bitidx in sorted(list(wbits)):
                # Are the names arbitrary? Lets just re-create
                line("bit %08X_%03u_%02u" % (frame, word, bitidx))

    for k, v in tags.items():
        line("tag %s %u" % (k, v))
